from ml4ir.base.model.scoring.scoring_base import ScoringBase


class PointwiseScoring(ScoringBase):
    def get_scoring_fn(self, **kwargs):
        """
        Define a pointwise ranking scoring function with specified architecture
        """

        def _scoring_fn(features):
            return self.architecture_op(features)

        return _scoring_fn



# Run from ml4ir/python directory as: python3 ml4ir/applications/ranking/data/scripts/create_dataset.py

import argparse
import os
import random
import socket
import time
import traceback
import datetime as dt
import pandas as pd
import numpy as np
from logging import Logger

from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.io import logging_utils
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import TFRecordTypeKey

# Defaults
DATA_DIR = "ml4ir/applications/ranking/data/"
OUT_DIR = "ml4ir/applications/ranking/data/synthetic/"
FEATURE_CONFIG = "ml4ir/applications/ranking/data/feature_config.yaml"
FEATURE_HIGHVAL = {"text_match_bool": [0, 1]}
FEATURE_NUM_RESULTS = "num_results"
MAX_NUM_RECORDS = 50
NUM_SAMPLES = 10
RANDOM_STATE = 123


def run_dataset_creation(
    data_dir: str = DATA_DIR,
    out_dir: str = OUT_DIR,
    feature_config_path: str = FEATURE_CONFIG,
    feature_highval: dict = FEATURE_HIGHVAL,
    feature_num_results: str = FEATURE_NUM_RESULTS,
    max_num_records: int = MAX_NUM_RECORDS,
    num_samples: int = NUM_SAMPLES,
    random_state: int = RANDOM_STATE,
):
    """
    1. Loads example data
    2. Builds specified synthetic data size by sampling from example data
    3. Adds catastrophic failures specifically
    4. For now, write out to CSV. In future could return df directly
    """
    # Setup logging
    file_io = LocalIO()
    logger: Logger = setup_logging(file_io)
    file_io.set_logger(logger)

    try:
        # Set seeds
        set_seeds(random_state)
        logger.info("Set seeds with initial random state {}".format(random_state))

        # Load and parse feature config
        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            feature_config_dict=file_io.read_yaml(feature_config_path),
            logger=logger,
        )
        logger.info("Feature config parsed and loaded")

        # Create output location
        file_io.make_directory(out_dir)
        out_file = os.path.join(
            out_dir, "synthetic_data_{}.csv".format(dt.datetime.now().strftime("%Y%m%d-%H%M%S"))
        )

        # Build data
        seed_data = load_seed_data(data_dir, logger, file_io)

        df_synthetic = fill_data(
            seed_data,
            max_num_records,
            feature_config,
            feature_highval,
            feature_num_results,
            num_samples,
            logger,
        )
        file_io.write_df(df_synthetic, outfile=out_file, index=False)
        logger.info("Synthetic data created! Location: {}".format(out_file))
        return df_synthetic

    except Exception as e:
        logger.error("!!! Error creating synthetic data: !!!\n{}".format(str(e)))
        traceback.print_exc()
        return


def setup_logging(file_io: LocalIO):
    run_id = "-".join([socket.gethostname(), time.strftime("%Y%m%d-%H%M%S")])
    logs_dir: str = os.path.join("logs", run_id)
    file_io.make_directory(logs_dir, clear_dir=True)

    outfile: str = os.path.join(logs_dir, "output_log.csv")
    logger = logging_utils.setup_logging(reset=True, file_name=outfile, log_to_file=True)

    logger.info("Logging initialized. Saving logs to : {}".format(logs_dir))
    logger.info("Run ID: {}".format(run_id))
    return logger


def set_seeds(random_state):
    # for repeatability
    np.random.seed(random_state)
    random.seed(random_state)


def load_seed_data(data_dir, logger, file_io: LocalIO):
    # Start with loading in CSV format only
    if file_io.path_exists(data_dir):
        dfs = file_io.get_files_in_directory(data_dir)
        return file_io.read_df_list(dfs)
    else:
        logger.error("Error! Data directory must exist and be specified")


def generate_key(seed_dict, query_key, logger):
    seed_id = str(random.sample(seed_dict[query_key], 1)[0])
    q_id = "".join(random.sample(seed_id, len(seed_id)))
    if q_id in set(seed_dict.keys()):
        logger.warning(
            "Generated query key is duplicate, regenerating. If happens repeatedly, query key may be too short"
        )
        return generate_key(seed_dict, query_key, logger)
    return q_id


def filter_nres(seed_data, feature_query_key, feature_rank, feature_nres, max_num_records, logger):
    nres = "num_results_calc"
    if nres in set(seed_data.columns):
        nres = "{}_seed".format(nres)
    seed_data[nres] = seed_data.groupby(feature_query_key).transform("count")[feature_rank]

    # If the number of results feature name was given, remove queries with not all results included
    if feature_nres and feature_nres in set(seed_data.columns):
        seed_data = seed_data.loc[
            (seed_data[feature_nres].dropna().astype("int64") == seed_data[nres])
        ]
    # Filter to max number of records if given
    if max_num_records:
        seed_data = seed_data.loc[seed_data[nres] <= max_num_records]

    if len(seed_data) == 0:
        logger.error("All seed data has been filtered out!")
    return seed_data, nres


def add_feature_highval(row_dict, feature_highval, is_clicked, click_rank):
    # Create ranking-specific high value scenario
    # of the format {high_value_feature_1:[label1_val, label2_val]}
    # Not creating examples of non-catastrophic failure, ie where the clicked record is not a name match
    for hvk in feature_highval.keys():
        if is_clicked and click_rank > 1:
            row_dict[hvk] = feature_highval[hvk][1]
        else:
            row_dict[hvk] = feature_highval[hvk][0]
    return row_dict


def fill_data(
    seed_data, max_num_records, feature_config, feature_highval, feature_nres, num_samples, logger
):
    """
    Creates synthetic data using source data as template sampling source
    Fastest way is to create a list of dicts, then create the DF from the list
    For the specified number of samples:
    1. Generate a unique query key
    2. Random sample (with replacement) number of results from seed data
    3. Random sample (with replacement) to fill each result from seed data
    """

    # Set up new DF and seed data distributions for sampling
    rows_df = []
    seed_dict = (
        {}
    )  # the keys are seed data features, the values are lists of associated values in seed data
    seed_dict_clicked = {}
    seed_dict_unclicked = {}

    # Get names of essential features from feature_config
    feature_query_key = feature_config.get_query_key("name")
    feature_rank = feature_config.get_rank("name")
    feature_label = feature_config.get_label("name")

    # For ranking data, calculate number of results from data and filter accordingly
    seed_data, nres = filter_nres(
        seed_data, feature_query_key, feature_rank, feature_nres, max_num_records, logger
    )

    # For ranking data, get rank of clicked result
    seed_data = pd.merge(
        seed_data,
        seed_data[seed_data[feature_label] == True][
            [feature_rank, feature_query_key]
        ].rename(  # noqa: E712 # FIXME
            columns={feature_rank: "click_rank"}
        ),
        on=feature_query_key,
        how="left",
    )

    # Now generate data distributions from seed data
    for mycol in list(seed_data.columns):
        seed_dict[mycol] = list(seed_data[mycol].values)
        seed_dict_clicked[mycol] = list(seed_data[seed_data[feature_label] == 1][mycol].values)
        seed_dict_unclicked[mycol] = list(seed_data[seed_data[feature_label] == 0][mycol].values)

    for _ in range(num_samples):
        # Generate a synthetic query key
        q_id = generate_key(seed_dict, feature_query_key, logger)

        # Sample to choose the number of members of the query key (in ranking, the number of results)
        q_nseq = random.sample(seed_dict[nres], 1)[0]

        # Sample to choose which result is clicked, sampling only from query keys with the same nseq (number of results)
        click_rank = random.sample(
            list(seed_data[seed_data[nres] == q_nseq].click_rank.values), 1
        )[0]

        # Build characteristics for each member of the query key (in ranking, each result of the query)
        for i in range(int(q_nseq)):
            is_clicked = i + 1 == click_rank
            row_dict = {
                feature_query_key: q_id,
                nres: q_nseq,
                "click_rank": click_rank,
                feature_rank: i + 1,
                feature_label: int(is_clicked),
            }
            for mycol in list(seed_data.columns):
                # Could also let user specify distribution to sample from for each feature
                if mycol not in row_dict.keys() and is_clicked:
                    # row_dict[mycol] = random.sample(seed_dict_clicked[mycol])
                    row_dict[
                        mycol
                    ] = None  # Default value, check type # Could also be functions we pass in
                elif mycol not in row_dict.keys() and not is_clicked:
                    # row_dict[mycol] = random.sample(seed_dict_unclicked[mycol])
                    row_dict[mycol] = None

            row_dict = add_feature_highval(row_dict, feature_highval, is_clicked, click_rank)

            rows_df.append(row_dict)
    return pd.DataFrame(rows_df)


def main(args):

    run_dataset_creation(
        args.data_dir,
        args.out_dir,
        args.feature_config,
        args.feature_highval,
        args.feature_num_results,
        args.max_num_records,
        args.num_samples,
        args.random_state,
    )


if __name__ == "__main__":
    # Define args
    parser = argparse.ArgumentParser()
    parser.add_argument("-data_dir", default=DATA_DIR)
    parser.add_argument("-out_dir", default=OUT_DIR)
    parser.add_argument("-feature_config", default=FEATURE_CONFIG)
    parser.add_argument("-feature_highval", default=FEATURE_HIGHVAL)
    parser.add_argument("-feature_num_results", default=FEATURE_NUM_RESULTS)
    parser.add_argument("-max_num_records", default=MAX_NUM_RECORDS)
    parser.add_argument("-num_samples", default=NUM_SAMPLES)
    parser.add_argument("-random_state", default=RANDOM_STATE)
    args = parser.parse_args()
    main(args)



# Stating loss functions related to classification use-case.
from tensorflow.keras import layers
from tensorflow.keras import losses

from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.applications.ranking.config.keys import LossKey


def get_loss(loss_key) -> RelevanceLossBase:
    """
    Factory to get relevance loss related to classification use-case.

    Parameters
    ----------
    loss_key : str
        LossKey name

    Returns
    -------
    RelevanceLossBase
        Corresponding loss object
    """
    if loss_key == LossKey.CATEGORICAL_CROSS_ENTROPY:
        return CategoricalCrossEntropy()
    else:
        raise NotImplementedError


class CategoricalCrossEntropy(RelevanceLossBase):
    def get_loss_fn(self, **kwargs):
        """
        Define a categorical cross entropy loss

        Returns
        -------
        function
            Categorical cross entropy loss function
        """
        cce = losses.CategoricalCrossentropy(reduction=losses.Reduction.SUM_OVER_BATCH_SIZE)

        def _loss_fn(y_true, y_pred):
            return cce(y_true, y_pred)

        return _loss_fn

    def get_final_activation_op(self, output_name):
        """
        Define softmax activation function

        Parameters
        ----------
        output_name : str
            Name of the output to use for final activation layer

        Returns
        -------
        function
            Softmax activation function
        """
        return lambda logits, mask: layers.Activation("softmax", name=output_name)(logits)



import sys
import ast
from argparse import Namespace
import tensorflow as tf
from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer
import pandas as pd
import pathlib
from scipy import stats

from ml4ir.base.pipeline import RelevancePipeline
from ml4ir.base.config.keys import ArchitectureKey
from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.model.scoring.scoring_model import ScorerBase, RelevanceScorer
from ml4ir.base.model.scoring.interaction_model import InteractionModel, UnivariateInteractionModel
from ml4ir.base.model.optimizers.optimizer import get_optimizer
from ml4ir.applications.ranking.model.ranking_model import RankingModel, LinearRankingModel, RankingConstants
from ml4ir.applications.ranking.config.keys import LossKey
from ml4ir.applications.ranking.config.keys import MetricKey
from ml4ir.applications.ranking.config.keys import ScoringTypeKey
from ml4ir.applications.ranking.model.losses import loss_factory
from ml4ir.applications.ranking.model.metrics import metric_factory
from ml4ir.applications.ranking.config.parse_args import get_args


from typing import Union, List, Type


class RankingPipeline(RelevancePipeline):
    """Base class that defines a pipeline to train, evaluate and save
    a RankingModel using ml4ir"""

    def __init__(self, args: Namespace):
        """
        Constructor to create a RelevancePipeline object to train, evaluate
        and save a model on ml4ir.
        This method sets up data, logs, models directories, file handlers used.
        The method also loads and sets up the FeatureConfig for the model training
        pipeline

        Parameters
        ----------
        args: argparse Namespace
            arguments to be used with the pipeline.
            Typically, passed from command line arguments
        """
        self.scoring_type = args.scoring_type
        self.loss_type = args.loss_type

        super().__init__(args)

        if self.model_config["architecture_key"] == ArchitectureKey.LINEAR:
            self.ranking_model_cls = LinearRankingModel
        else:
            self.ranking_model_cls = RankingModel

    def get_relevance_model(self, feature_layer_keys_to_fns={}) -> RelevanceModel:
        """
        Creates a RankingModel that can be used for training and evaluating

        Parameters
        ----------
        feature_layer_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to tensorflow compatible
            function definitions that can now be used in the InteractionModel
            as a feature function to transform input features

        Returns
        -------
        `RankingModel`
            RankingModel that can be used for training and evaluating
            a ranking model

        Notes
        -----
        Override this method to create custom loss, scorer, model objects
        """

        # Define interaction model
        interaction_model: InteractionModel = UnivariateInteractionModel(
            feature_config=self.feature_config,
            feature_layer_keys_to_fns=feature_layer_keys_to_fns,
            tfrecord_type=self.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            file_io=self.file_io,
        )

        # Define loss object from loss key
        loss: RelevanceLossBase = loss_factory.get_loss(
            loss_key=self.loss_key, scoring_type=self.scoring_type
        )

        # Define scorer
        scorer: ScorerBase = RelevanceScorer(
            feature_config=self.feature_config,
            model_config=self.model_config,
            interaction_model=interaction_model,
            loss=loss,
            output_name=self.args.output_name,
            logger=self.logger,
            file_io=self.file_io,
        )

        # Define metrics objects from metrics keys
        metrics: List[Union[Type[Metric], str]] = [
            metric_factory.get_metric(metric_key=metric_key) for metric_key in self.metrics_keys
        ]

        optimizer: Optimizer = get_optimizer(model_config=self.model_config)

        # Combine the above to define a RelevanceModel
        relevance_model: RelevanceModel = self.ranking_model_cls(
            feature_config=self.feature_config,
            tfrecord_type=self.tfrecord_type,
            scorer=scorer,
            metrics=metrics,
            optimizer=optimizer,
            model_file=self.model_file,
            initialize_layers_dict=ast.literal_eval(self.args.initialize_layers_dict),
            freeze_layers_list=ast.literal_eval(self.args.freeze_layers_list),
            compile_keras_model=self.args.compile_keras_model,
            output_name=self.args.output_name,
            file_io=self.local_io,
            logger=self.logger,
        )

        return relevance_model

    def validate_args(self):
        """
        Validate the arguments to be used with RelevancePipeline
        """
        super().validate_args()

        if self.loss_key not in LossKey.get_all_keys():
            raise Exception(
                "Loss specified [{}] is not one of : {}".format(
                    self.loss_key, LossKey.get_all_keys()
                )
            )

        for metric_key in self.metrics_keys:
            if metric_key not in MetricKey.get_all_keys():
                raise Exception(
                    "Metric specified [{}] is not one of : {}".format(
                        metric_key, MetricKey.get_all_keys()
                    )
                )

        if self.scoring_type not in ScoringTypeKey.get_all_keys():
            raise Exception(
                "Scoring type specified [{}] is not one of : {}".format(
                    self.scoring_type, ScoringTypeKey.get_all_keys()
                )
            )

    def create_pipeline_for_kfold(self, args):
        """
        Create a RankingPipeline object used in running kfold cross validation.
        """
        return RankingPipeline(args=args)

    def kfold_analysis(self, base_logs_dir, run_id, num_folds, pvalue_threshold=0.1, metrics=None):
        """
        Aggregate results of the k-fold runs and perform t-test on the results between old(prod model) and
        new model's w.r.t the specified metrics.

        Parameters
        ----------
        base_logs_dir : int
            Total number of folds
        run_id : int
            current fold number
        num_folds: int
            Total number of folds
        pvalue_threshold: float
            the threshold used for pvalue to assess significance
        metrics: list
            List of metrics to include in the kfold analysis
        """
        if not metrics:
            return
        metrics_formated = []
        datasets = ['train', 'val', 'test']
        for dataset in datasets:
            for metric in metrics:
                metrics_formated.append("{}_old_{}".format(dataset, metric))
                metrics_formated.append("{}_new_{}".format(dataset, metric))
        rows = []
        for i in range(num_folds):
            row = {"fold": i}
            logs_dir = pathlib.Path(base_logs_dir) / run_id / "fold_{}".format(i) / run_id
            with open(logs_dir / "_SUCCESS", 'r') as f:
                for line in f:
                    parts = line.split(',')
                    if parts[0] in metrics_formated:
                        row[parts[0]] = float(parts[1])
            rows.append(row)
        results = pd.DataFrame(rows)
        self.logger.info(results)
        ttest_result = []
        # calculate statistical paired t-test
        for metric in metrics:
            for dataset in ['train', 'val', 'test']:
                t_test_stat, pvalue = stats.ttest_rel(results["{}_old_{}".format(dataset, metric)],
                                                      results["{}_new_{}".format(dataset, metric)])
                ttest = {"metric":metric, "dataset": dataset, "t_test_stat": t_test_stat, "p_value": pvalue,
                         "is_stat_sig": pvalue < pvalue_threshold}
                self.logger.info("{}_t_test_stat={}".format(dataset, t_test_stat))
                self.logger.info(
                    "{}_pvalue={} --> (statistically significant={})".format(dataset, pvalue, pvalue < pvalue_threshold))
                ttest_result.append(ttest)
        return pd.DataFrame(ttest_result)

    def run_kfold_analysis(self, logs_dir, run_id, num_folds, metrics):
        """
        Running the kfold analysis for ranking.

        Parameters:
        -----------
        logs_dir: str
            path to logs directory
        run_id: str
            string run_id
        num_folds: int
            number of folds
        metrics: list
            list of metrics to include in the kfold analysis

        Returns:
        --------
        summary of the kfold analysis
        """
        #TODO:Choose the best model among the folds

        return str(self.kfold_analysis(logs_dir, run_id, num_folds, RankingConstants.TTEST_PVALUE_THRESHOLD,
                               metrics))


def main(argv):
    # Define args
    args: Namespace = get_args(argv)
    # Initialize Relevance Pipeline and run in train/inference mode
    rp = RankingPipeline(args=args)
    rp.run()


if __name__ == "__main__":
    main(sys.argv[1:])



import tensorflow as tf
from tensorflow import io
from tensorflow import data
from tensorflow import sparse
from tensorflow import image
from logging import Logger

from ml4ir.base.io.file_io import FileIO
from ml4ir.base.features.preprocessing import PreprocessingMap
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import SequenceExampleTypeKey, TFRecordTypeKey

from typing import Optional


class TFRecordParser(object):
    """
    Base class for parsing TFRecord examples. This class consolidates the
    parsing and feature extraction pipeline for both Example and SequenceExample
    protobuf messages
    """

    def __init__(
        self,
        feature_config: FeatureConfig,
        preprocessing_map: PreprocessingMap,
        required_fields_only: Optional[bool] = False,
    ):
        """
        Constructor method for instantiating a TFRecordParser object

        Parameters
        ----------
        feature_config : `FeatureConfig`
            FeatureConfig object defining context and sequence feature information
        preprocessing_map : `PreprocessingMap` object
            Object mapping preprocessing feature function names to their definitons
        required_fields_only : bool, optional
            Whether to only use required fields from the feature_config
        """
        self.feature_config = feature_config
        self.preprocessing_map = preprocessing_map
        self.required_fields_only = required_fields_only
        self.features_spec = self.get_features_spec()

    def get_features_spec(self):
        """
        Define the features spec from the feature_config.
        The features spec will be used to parse the serialized TFRecord

        Returns
        -------
        dict
            feature specification dictionary that can be used to parse TFRecords

        Notes
        -----
        For SequenceExample messages, this method returns a pair of dictionaries,
        one each for context and sequence features.
        """
        raise NotImplementedError

    def extract_features_from_proto(self, proto):
        """
        Parse the serialized proto string to extract features

        Parameters
        ----------
        proto: tf.Tensor
            A scalar string tensor that is the serialized form of a TFRecord object

        Returns
        -------
        dict of Tensors
            Dictionary of features extracted from the proto as per the features_spec

        Notes
        -----
        For SequenceExample proto messages, this function returns two dictionaries,
        one for context and another for sequence feature tensors.
        For Example proto messages, this function returns a single dictionary of feature tensors.
        """
        raise NotImplementedError

    def get_default_tensor(self, feature_info, sequence_size=0):
        """
        Get the default tensor for a given feature configuration

        Parameters
        ----------
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        sequence_size: int, optional
            Number of elements in the sequence of a SequenceExample

        Returns
        -------
        tf.Tensor
            Tensor object that can be used as a default tensor if the expected feature
            is missing from the TFRecord
        """
        raise NotImplementedError

    def get_feature(self, feature_info, extracted_features, sequence_size=0):
        """
        Fetch the feature from the feature dictionary of extracted features

        Parameters
        ----------
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        extracted_features: dict
            Dictionary of feature tensors extracted by parsing the serialized TFRecord
        sequence_size: int, optional
            Number of elements in the sequence of a SequenceExample

        Returns
        -------
        tf.Tensor
            Feature tensor that is obtained from the extracted features for the given
            feature_info
        """
        raise NotImplementedError

    def generate_and_add_mask(self, extracted_features, features_dict):
        """
        Create a mask to identify padded values

        Parameters
        ----------
        extracted_features: dict
            Dictionary of tensors extracted from the serialized TFRecord
        features_dict: dict
            Dictionary of tensors that will be used for model training/serving
            as inputs to the model

        Returns
        -------
        features_dict: dict
            Dictionary of tensors that will be used for model training/serving updated
            with the mask tensor if applicable
        sequence_size: int
            Number of elements in the sequence of the TFRecord
        """
        raise NotImplementedError

    def pad_feature(self, feature_tensor, feature_info):
        """
        Pad the feature to the `max_sequence_size` in order to create
        uniform data batches for training
        Parameters
        ----------
        feature_tensor: tf.Tensor
            Feature tensor to be padded
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        Returns
        -------
        tf.Tensor
            Feature tensor padded to the `max_sequence_size`
        """

        raise NotImplementedError

    def preprocess_feature(self, feature_tensor, feature_info):
        """
        Preprocess feature based on the feature configuration

        Parameters
        ----------
        feature_tensor: `tf.Tensor`
            input feature tensor to be preprocessed
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config

        Returns
        -------
        `tf.Tensor`
            preprocessed tensor object

        Notes
        -----
        Only preprocessing functions part of the `preprocessing_map`
        can be used in this function for preprocessing at data loading

        Pass custom preprocessing functions while instantiating the
        RelevanceDataset object with `preprocessing_keys_to_fns` argument
        """
        preprocessing_info = feature_info.get("preprocessing_info")

        if preprocessing_info:
            for preprocessing_step in preprocessing_info:
                preprocessing_fn = self.preprocessing_map.get_fn(preprocessing_step["fn"])
                if preprocessing_fn:
                    feature_tensor = preprocessing_fn(
                        feature_tensor, **preprocessing_step.get("args", {})
                    )

        return feature_tensor

    def get_parse_fn(self) -> tf.function:
        """
        Define a parsing function that will be used to load the TFRecordDataset
        and create input features for the model.

        Returns
        -------
        `tf.function`
            Parsing function that takes in a serialized TFRecord protobuf
            message and extracts a dictionary of feature tensors

        Notes
        -----
        This function will also be used with the TFRecord serving signature in the
        saved model.
        """

        @tf.function
        def _parse_fn(proto):
            """
            Parse a serialized TFRecord proto message

            Parameters
            ----------
            proto: tf.Tensor
                Scalar string tensor that needs to be parsed to extract features

            Returns
            -------
            features: dict
                Dictionary of feature tensors
            labels: tf.Tensor
                Label feature tensor used for training
            """
            # Parse the proto message to extract feature tensors
            extracted_features = self.extract_features_from_proto(proto)
            features_dict = dict()

            # Create a mask tensor and add to the features dictionary
            features_dict, sequence_size = self.generate_and_add_mask(
                extracted_features, features_dict
            )

            # Process all features, including label to construct the feature tensor dictionary
            for feature_info in self.feature_config.get_all_features(include_mask=False):
                feature_node_name = feature_info.get("node_name", feature_info["name"])

                # Fetch the feature corresponding to the feature_info from the extracted features
                feature_tensor = self.get_feature(feature_info, extracted_features, sequence_size)

                # Pad the extracted feature to the max_sequence_size for training
                feature_tensor = self.pad_feature(feature_tensor, feature_info)

                # Preprocess the extracted feature using the specification from
                # the FeatureConfig and functions from PreprocessingMap
                feature_tensor = self.preprocess_feature(feature_tensor, feature_info)

                # Add the processed feature tensor to the features dictionary
                features_dict[feature_node_name] = feature_tensor

            # Extract the label feature to return separately
            labels = features_dict.pop(self.feature_config.get_label(key="name"))

            # return X and y which can be used with fit(), predict() and evaluate()
            return features_dict, labels

        return _parse_fn


class TFRecordExampleParser(TFRecordParser):
    """
    Class for parsing Example TFRecord protobuf messages
    """

    def get_features_spec(self):
        """
        Define the features spec from the feature_config.
        This will be used to parse the serialized TFRecord

        Returns
        -------
        dict
            feature specification dictionary that can be used to parse TFRecords
        """
        features_spec = dict()

        for feature_info in self.feature_config.get_all_features():
            serving_info = feature_info["serving_info"]
            if not self.required_fields_only or serving_info.get(
                "required", feature_info["trainable"]
            ):
                feature_name = feature_info["name"]
                dtype = feature_info["dtype"]
                default_value = self.feature_config.get_default_value(feature_info)
                features_spec[feature_name] = io.FixedLenFeature([], dtype, default_value=default_value)

        return features_spec

    def extract_features_from_proto(self, serialized):
        """
        Parse the serialized proto string to extract features

        Parameters
        ----------
        proto: tf.Tensor
            A scalar string tensor that is the serialized form of a TFRecord object

        Returns
        -------
        dict of Tensors
            Dictionary of features extracted from the proto as per the features_spec
        """
        return io.parse_single_example(serialized=serialized, features=self.features_spec)

    def get_default_tensor(self, feature_info, sequence_size=0):
        """
        Get the default tensor for a given feature configuration

        Parameters
        ----------
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        sequence_size: int, optional
            Number of elements in the sequence of a SequenceExample

        Returns
        -------
        tf.Tensor
            Tensor object that can be used as a default tensor if the expected feature
            is missing from the TFRecord
        """
        return tf.constant(
            value=self.feature_config.get_default_value(feature_info), dtype=feature_info["dtype"],
        )

    def get_feature(self, feature_info, extracted_features, sequence_size=0):
        """
        Fetch the feature from the feature dictionary of extracted features

        Parameters
        ----------
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        extracted_features: dict
            Dictionary of feature tensors extracted by parsing the serialized TFRecord
        sequence_size: int, optional
            Number of elements in the sequence of a SequenceExample

        Returns
        -------
        tf.Tensor
            Feature tensor that is obtained from the extracted features for the given
            feature_info
        """
        default_tensor = self.get_default_tensor(feature_info, sequence_size)

        feature_tensor = extracted_features.get(feature_info["name"], default_tensor)
        
        # Adjust shape
        feature_tensor = tf.expand_dims(feature_tensor, axis=0)

        return feature_tensor

    def generate_and_add_mask(self, extracted_features, features_dict):
        """
        Create a mask to identify padded values

        Parameters
        ----------
        extracted_features: dict
            Dictionary of tensors extracted from the serialized TFRecord
        features_dict: dict
            Dictionary of tensors that will be used for model training/serving
            as inputs to the model

        Returns
        -------
        features_dict: dict
            Dictionary of tensors that will be used for model training/serving updated
            with the mask tensor if applicable
        sequence_size: int
            Number of elements in the sequence of the TFRecord
        """
        return features_dict, tf.constant(0)

    def pad_feature(self, feature_tensor, feature_info):
        """
        Pad the feature to the `max_sequence_size` in order to create
        uniform data batches for training
        Parameters
        ----------
        feature_tensor: tf.Tensor
            Feature tensor to be padded
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        Returns
        -------
        tf.Tensor
            Feature tensor padded to the `max_sequence_size`
        """
        return feature_tensor


class TFRecordSequenceExampleParser(TFRecordParser):
    def __init__(
        self,
        feature_config: FeatureConfig,
        preprocessing_map: PreprocessingMap,
        required_fields_only: Optional[bool] = False,
        pad_sequence: Optional[bool] = True,
        max_sequence_size: Optional[int] = 25,
    ):
        """
        Constructor method for instantiating a TFRecordParser object

        Parameters
        ----------
        feature_config : `FeatureConfig`
            FeatureConfig object defining context and sequence feature information
        preprocessing_map : `PreprocessingMap` object
            Object mapping preprocessing feature function names to their definitons
        required_fields_only : bool, optional
            Whether to only use required fields from the feature_config
        pad_sequence: bool, optional
            Whether to pad sequence
        max_sequence_size: int, optional
            Maximum number of sequence per query. Used for padding
        """
        self.pad_sequence = pad_sequence
        self.max_sequence_size = max_sequence_size
        super(TFRecordSequenceExampleParser, self).__init__(
            feature_config=feature_config,
            preprocessing_map=preprocessing_map,
            required_fields_only=required_fields_only,
        )

    def get_features_spec(self):
        """
        Define the features spec from the feature_config.
        This will be used to parse the serialized TFRecord

        Returns
        -------
        dict
            Feature specification dictionary that can be used to parse
            Context features from the serialized SequenceExample
        dict
            Feature specification dictionary that can be used to parse
            Sequence features (or feature lists) from the serialized SequenceExample
        """

        context_features_spec = dict()
        sequence_features_spec = dict()

        for feature_info in self.feature_config.get_all_features():
            if feature_info.get("name") == self.feature_config.get_mask("name"):
                continue
            serving_info = feature_info["serving_info"]
            if not self.required_fields_only or serving_info.get("required", feature_info["trainable"]):
                feature_name = feature_info["name"]
                dtype = feature_info["dtype"]
                default_value = self.feature_config.get_default_value(
                    feature_info)
                if feature_info["tfrecord_type"] == SequenceExampleTypeKey.CONTEXT:
                    context_features_spec[feature_name] = io.FixedLenFeature(
                        [], dtype, default_value=default_value
                    )
                elif feature_info["tfrecord_type"] == SequenceExampleTypeKey.SEQUENCE:
                    sequence_features_spec[feature_name] = io.VarLenFeature(
                        dtype=dtype)
                else:
                    raise KeyError("Invalid SequenceExample type: {}".format(
                        feature_info["tfrecord_type"]))

        return context_features_spec, sequence_features_spec

    def extract_features_from_proto(self, serialized):
        """
        Parse the serialized proto string to extract features

        Parameters
        ----------
        proto: tf.Tensor
            A scalar string tensor that is the serialized form of a TFRecord object

        Returns
        -------
        dict of Tensors
            Dictionary of context feature tensors extracted from the proto
            as per the `features_spec`
        dict of Tensors
            Dictionary of sequence feature tensors extracted from the proto
            as per the `features_spec`
        """
        return io.parse_single_sequence_example(
            serialized=serialized,
            context_features=self.features_spec[0],
            sequence_features=self.features_spec[1],
        )

    def get_default_tensor(self, feature_info, sequence_size):
        """
        Get the default tensor for a given feature configuration
        Parameters
        ----------
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        sequence_size: int, optional
            Number of elements in the sequence of a SequenceExample
        Returns
        -------
        tf.Tensor
            Tensor object that can be used as a default tensor if the expected feature
            is missing from the TFRecord
        """
        if feature_info.get("tfrecord_type", SequenceExampleTypeKey.CONTEXT) == SequenceExampleTypeKey.CONTEXT:
            return tf.constant(
                value=self.feature_config.get_default_value(feature_info), dtype=feature_info["dtype"],
            )
        else:
            return tf.fill(
                value=tf.constant(
                    value=self.feature_config.get_default_value(feature_info),
                    dtype=feature_info["dtype"],
                ),
                dims=[sequence_size],
            )

    def get_feature(self, feature_info, extracted_features, sequence_size):
        """
        Fetch the feature from the feature dictionary of extracted features
        Parameters
        ----------
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        extracted_features: dict
            Dictionary of feature tensors extracted by parsing the serialized TFRecord
        sequence_size: int, optional
            Number of elements in the sequence of a SequenceExample
        Returns
        -------
        tf.Tensor
            Feature tensor that is obtained from the extracted features for the given
            feature_info
        """
        extracted_context_features, extracted_sequence_features = extracted_features

        default_tensor = self.get_default_tensor(feature_info, sequence_size)

        if feature_info["tfrecord_type"] == SequenceExampleTypeKey.CONTEXT:
            feature_tensor = extracted_context_features.get(
                feature_info["name"], default_tensor)
            # Adjust shape
            feature_tensor = tf.expand_dims(feature_tensor, axis=0)
        else:
            feature_tensor = extracted_sequence_features.get(
                feature_info["name"], default_tensor)
            if isinstance(feature_tensor, sparse.SparseTensor):
                feature_tensor = sparse.reset_shape(feature_tensor)
                feature_tensor = sparse.to_dense(feature_tensor)
                feature_tensor = tf.squeeze(feature_tensor, axis=0)

        return feature_tensor

    def generate_and_add_mask(self, extracted_features, features_dict):
        """
        Create a mask to identify padded values

        Parameters
        ----------
        extracted_features: dict
            Dictionary of tensors extracted from the serialized TFRecord
        features_dict: dict
            Dictionary of tensors that will be used for model training/serving
            as inputs to the model

        Returns
        -------
        features_dict: dict
            Dictionary of tensors that will be used for model training/serving updated
            with the mask tensor if applicable
        sequence_size: int
            Number of elements in the sequence of the TFRecord
        """
        context_features, sequence_features = extracted_features
        if (
            self.required_fields_only
            and not self.feature_config.get_rank("serving_info")["required"]
        ):
            """
            Define dummy mask if the rank field is not a required field for serving
            NOTE:
            This masks all max_sequence_size as 1 as there is no real way to know
            the number of sequence in the query. There is no predefined required field,
            and hence we would need to do a full pass of all features to find the record shape.
            This approach might be unstable if different features have different shapes.
            Hence we just mask all sequence
            """
            mask = tf.constant(
                value=1,
                shape=[self.max_sequence_size],
                dtype=self.feature_config.get_rank("dtype"),
            )
            sequence_size = tf.constant(self.max_sequence_size, dtype=tf.int64)
        else:
            # Typically used at training time, to pad/clip to a fixed number of sequence per query

            # Use rank as a reference tensor to infer shape/sequence_size in query
            reference_tensor = sequence_features.get(self.feature_config.get_rank(key="node_name"))

            # Add mask for identifying padded sequence
            mask = tf.ones_like(sparse.to_dense(sparse.reset_shape(reference_tensor)))

            if self.pad_sequence:
                mask = tf.squeeze(mask, axis=0)

                def crop_fn():
                    # NOTE: We currently ignore these cases as there is no clear
                    # way to select max_sequence_size from all the sequence features
                    tf.print("\n[WARN] Bad query found. Number of sequence : ", tf.shape(mask)[0])
                    return mask

                mask = tf.cond(
                    tf.shape(mask)[0] <= self.max_sequence_size,
                    # Pad if there are missing sequence
                    lambda: tf.pad(
                        mask, [[0, self.max_sequence_size - tf.shape(mask)[0]]]),
                    # Crop if there are extra sequence
                    crop_fn,
                )
                sequence_size = tf.constant(self.max_sequence_size, dtype=tf.int64)
            else:
                mask = tf.squeeze(mask, axis=0)
                sequence_size = tf.cast(tf.reduce_sum(mask), tf.int64)

        # Check validity of mask
        tf.debugging.assert_greater(sequence_size, tf.constant(0, dtype=tf.int64))

        # Update features dictionary with the computed mask tensor
        features_dict["mask"] = mask

        return features_dict, sequence_size

    def pad_feature(self, feature_tensor, feature_info):
        """
        Pad the feature to the `max_sequence_size` in order to create
        uniform data batches for training
        Parameters
        ----------
        feature_tensor: tf.Tensor
            Feature tensor to be padded
        feature_info: dict
            Feature configuration information for the feature as specified in the feature_config
        Returns
        -------
        tf.Tensor
            Feature tensor padded to the `max_sequence_size`
        """
        if self.pad_sequence and feature_info["tfrecord_type"] == SequenceExampleTypeKey.SEQUENCE:
            pad_len = self.max_sequence_size - tf.shape(feature_tensor)[0]
            feature_tensor = tf.pad(feature_tensor, [[0, pad_len]])

        return feature_tensor


def get_parse_fn(
    tfrecord_type: str,
    feature_config: FeatureConfig,
    preprocessing_keys_to_fns: dict,
    max_sequence_size: int = 0,
    required_fields_only: bool = False,
    pad_sequence: bool = True,
) -> tf.function:
    """
    Create a parsing function to extract features from serialized TFRecord data
    using the definition from the FeatureConfig

    Parameters
    ----------
    tfrecord_type: {"example", "sequence_example"}
        Type of TFRecord data to be loaded into a dataset
    feature_config: `FeatureConfig` object
        FeatureConfig object defining the features to be extracted
    preprocessing_keys_to_fns: dict of(str, function), optional
        dictionary of function names mapped to function definitions
        that can now be used for preprocessing while loading the
        TFRecordDataset to create the RelevanceDataset object
    max_sequence_size: int
        Maximum number of sequence per query. Used for padding
    required_fields_only: bool, optional
        Whether to only use required fields from the feature_config
    pad_sequence: bool
        Whether to pad sequence

    Returns
    -------
    `tf.function`
        Parsing function that takes in a serialized SequenceExample or Example message
        and extracts a dictionary of feature tensors
    """
    # Define preprocessing functions
    preprocessing_map = PreprocessingMap()
    preprocessing_map.add_fns(preprocessing_keys_to_fns)

    # Generate parsing function
    if tfrecord_type == TFRecordTypeKey.EXAMPLE:
        parser: TFRecordParser = TFRecordExampleParser(
            feature_config=feature_config,
            preprocessing_map=preprocessing_map,
            required_fields_only=required_fields_only,
        )
    elif tfrecord_type == TFRecordTypeKey.SEQUENCE_EXAMPLE:
        parser = TFRecordSequenceExampleParser(
            feature_config=feature_config,
            preprocessing_map=preprocessing_map,
            max_sequence_size=max_sequence_size,
            required_fields_only=required_fields_only,
            pad_sequence=pad_sequence,
        )
    else:
        raise KeyError("Invalid TFRecord type specified: {}".format(tfrecord_type))

    return parser.get_parse_fn()


def read(
    data_dir: str,
    feature_config: FeatureConfig,
    tfrecord_type: str,
    file_io: FileIO,
    max_sequence_size: int = 0,
    batch_size: int = 0,
    preprocessing_keys_to_fns: dict = {},
    parse_tfrecord: bool = True,
    use_part_files: bool = False,
    logger: Logger = None,
    **kwargs
) -> data.TFRecordDataset:
    """
    Extract features by reading and parsing TFRecord data
    and converting into a TFRecordDataset using the FeatureConfig

    Parameters
    ----------
    data_dir: str
        path to the directory containing train, validation and test data
    feature_config: `FeatureConfig` object
        FeatureConfig object that defines the features to be loaded in the dataset
        and the preprocessing functions to be applied to each of them
    tfrecord_type: {"example", "sequence_example"}
        Type of the TFRecord protobuf message to be used for TFRecordDataset
    file_io: `FileIO` object
        file I/O handler objects for reading and writing data
    max_sequence_size: int, optional
        maximum number of sequence to be used with a single SequenceExample proto message
        The data will be appropriately padded or clipped to fit the max value specified
    batch_size: int, optional
        size of each data batch
    preprocessing_keys_to_fns: dict of(str, function), optional
        dictionary of function names mapped to function definitions
        that can now be used for preprocessing while loading the
        TFRecordDataset to create the RelevanceDataset object
    use_part_files: bool, optional
        load dataset from part files checked using "part-" prefix
    parse_tfrecord: bool, optional
        parse the TFRecord string from the dataset;
        returns strings as is otherwise
    logger: `Logger`, optional
        logging handler for status messages

    Returns
    -------
    `TFRecordDataset`
        TFRecordDataset loaded from the `data_dir` specified using the FeatureConfig
    """
    parse_fn = get_parse_fn(
        feature_config=feature_config,
        tfrecord_type=tfrecord_type,
        preprocessing_keys_to_fns=preprocessing_keys_to_fns,
        max_sequence_size=max_sequence_size,
    )

    # Get all tfrecord files in directory
    tfrecord_files = file_io.get_files_in_directory(
        data_dir,
        extension="" if use_part_files else ".tfrecord",
        prefix="part-" if use_part_files else "",
    )

    # Parse the protobuf data to create a TFRecordDataset
    dataset = data.TFRecordDataset(tfrecord_files)

    if parse_tfrecord:
        # Parallel calls set to AUTOTUNE: improved training performance by 40% with a classification model
        dataset = (
            dataset.map(parse_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)
            # .apply(data.experimental.ignore_errors())
        )

    # Create BatchedDataSet
    if batch_size:
        dataset = dataset.batch(batch_size, drop_remainder=False)

    if logger:
        logger.info(
            "Created TFRecordDataset from SequenceExample protobufs from {} files : {}".format(
                len(tfrecord_files), str(tfrecord_files)[:50]
            )
        )

    # We apply prefetch as it improved train/test/validation throughput by 30% in some real model training.
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

    return dataset



import os

from ml4ir.base.tests.test_base import RelevanceTestBase
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.features.feature_config import FeatureConfig


class RelevanceDatasetTest(RelevanceTestBase):
    def validate_dataset(self, ranking_dataset):
        # Check if the datasets were populated
        assert ranking_dataset.train is not None
        assert ranking_dataset.validation is not None
        assert ranking_dataset.test is not None

        # Check if data batch size is correct
        def check_shape(d):
            X, y = list(d)[0]
            assert y.shape[0] == self.args.batch_size
            for feature in X.keys():
                assert X[feature].shape[0] == self.args.batch_size

        map(
            check_shape,
            [
                ranking_dataset.train.take(1),
                ranking_dataset.validation.take(1),
                ranking_dataset.test.take(1),
            ],
        )

    def get_ranking_dataset(self, data_dir: str, data_format: str, feature_config_path: str):

        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=self.args.tfrecord_type,
            feature_config_dict=self.file_io.read_yaml(feature_config_path),
            logger=self.logger,
        )

        relevance_dataset = RelevanceDataset(
            data_dir=data_dir,
            data_format=data_format,
            feature_config=feature_config,
            tfrecord_type=self.args.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns={},
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.file_io,
            logger=self.logger,
        )

        return relevance_dataset

    def test_csv_dataset(self):
        """
        Test ranking dataset creation from CSV files
        using native tensorflow methods
        """

        data_dir = os.path.join(self.root_data_dir, "csv")
        feature_config_path = os.path.join(self.root_data_dir, "configs", self.feature_config_fname)

        ranking_dataset = self.get_ranking_dataset(
            data_dir=data_dir, data_format="csv", feature_config_path=feature_config_path
        )

        self.validate_dataset(ranking_dataset)

    def test_tfrecord_dataset(self):
        """
        Test ranking dataset creation from TFRecord files
        using pandas data load methods
        """

        data_dir = os.path.join(self.root_data_dir, "tfrecord")
        feature_config_path = os.path.join(self.root_data_dir, "configs", self.feature_config_fname)

        ranking_dataset = self.get_ranking_dataset(
            data_dir=data_dir, data_format="tfrecord", feature_config_path=feature_config_path
        )

        self.validate_dataset(ranking_dataset)



from ml4ir.base.config.keys import ScoringKey
from ml4ir.base.model.scoring.pointwise_scoring import PointwiseScoring
from ml4ir.base.model.scoring.pairwise_scoring import PairwiseScoring
from ml4ir.base.model.scoring.groupwise_scoring import GroupwiseScoring


def get_scoring_fn(scoring_key, model_config, loss_type):
    """
    Returns the scoring function based on the params
    """

    if scoring_key == ScoringKey.POINTWISE:
        scoring = PointwiseScoring(
            scoring_key=scoring_key, model_config=model_config, loss_type=loss_type
        )
    elif scoring_key == ScoringKey.PAIRWISE:
        scoring = PairwiseScoring(
            scoring_key=scoring_key, model_config=model_config, loss_type=loss_type
        )
    elif scoring_key == ScoringKey.GROUPWISE:
        scoring = GroupwiseScoring(
            scoring_key=scoring_key, model_config=model_config, loss_type=loss_type
        )
    else:
        raise NotImplementedError

    return scoring.get_scoring_fn()



from logging import Logger, INFO, DEBUG, ERROR
import pandas as pd

from typing import Optional


class FileIO(object):
    """Abstract class defining the file I/O handler methods"""

    def __init__(self, logger: Optional[Logger] = None):
        """
        Constructor method to create a FileIO handler object

        Parameters
        ----------
        logger : `Logger` object, optional
            logging handler object to instantiate FileIO object
            with the ability to log progress updates
        """
        self.logger = logger

    def set_logger(self, logger: Optional[Logger] = None):
        """
        Setter method to assign a logging handler to the FileIO object

        Parameters
        ----------
        logger : `Logger` object, optional
            logging handler object to be used with the FileIO object
            to log progress updates
        """
        self.logger = logger

    def log(self, string, mode=INFO):
        """
        Write specified string with preset logging object
        using the mode specified

        Parameters
        ----------
        string : str
            string text to be logged
        mode : int, optional
            One of the supported logging message types.
            Currently supported values are logging.INFO, DEBUG, ERROR
        """
        if self.logger:
            if mode == INFO:
                self.logger.info(string)
            elif mode == DEBUG:
                self.logger.info(string)
            elif mode == ERROR:
                self.logger.info(string)

    def make_directory(self, dir_path: str, clear_dir: bool = False) -> str:
        """
        Create directory structure specified recursively

        Parameters
        ----------
        dir_path : str
            path for directory to be create
        clear_dir : bool, optional
            clear contents on existing directory

        Returns
        -------
        str
            path to the directory created
        """
        raise NotImplementedError

    def read_df(
        self, infile: str, sep: str = ",", index_col: int = None
    ) -> Optional[pd.DataFrame]:
        """
        Load a pandas dataframe from a file

        Parameters
        ----------
        infile : str
            path to the csv input file
        sep : str, optional
            separator to use for loading file
        index_col : int, optional
            column to be used as index

        Returns
        -------
        `pandas.DataFrame`
            pandas dataframe loaded from specified path
        """
        raise NotImplementedError

    def read_df_list(self, infiles, sep=",", index_col=None) -> pd.DataFrame:
        """
        Load a pandas dataframe from a list of files by concatenating
        the individual dataframes from each file

        Parameters
        ----------
        infiles : list of str
            list of paths to the csv input files
        sep : str, optional
            separator to use for loading file
        index_col : int, optional
            column to be used as index

        Returns:
        `pandas.DataFrame`
            pandas dataframe loaded from specified path
        """
        raise NotImplementedError

    def write_df(self, df, outfile: str = None, sep: str = ",", index: bool = True):
        """
        Write a pandas dataframe to a file

        Parameters
        ----------
        df : `pandas.DataFrame`
            dataframe to be written
        outfile : str. optional
            path to the csv output file
        sep : str, optional
            separator to use for loading file
        index : bool, optional
            boolean specifying if index should be saved
        """
        raise NotImplementedError

    def read_text_file(self, infile) -> str:
        """
        Read text file and return as string

        Parameters
        ----------
        infile : str
            path to the text file

        Returns
        -------
        str
            file contents as a string
        """
        raise NotImplementedError

    def read_json(self, infile) -> dict:
        """
        Read JSON file and return a python dictionary

        Parameters
        ----------
        infile : str
            path to the json file

        Returns
        -------
        dict
            python dictionary loaded from JSON file
        """
        raise NotImplementedError

    def read_yaml(self, infile) -> dict:
        """
        Read YAML file and return a python dictionary

        Parameters
        ----------
        infile : str
            path to the YAML file

        Returns
        -------
        dict
            python dictionary loaded from JSON file
        """
        raise NotImplementedError

    def write_json(self, json_dict: dict, outfile: str):
        """
        Write dictionary to a JSON file

        Parameters
        ----------
        json_dict : dict
            dictionary to be dumped to json file
        outfile : str
            path to the output file
        """
        raise NotImplementedError

    def path_exists(self, path: str) -> bool:
        """
        Check if a file path exists

        Parameters
        ----------
        path : str
            check if path exists

        Returns
        -------
        bool
            True if path exists; False otherwise
        """
        raise NotImplementedError

    def get_files_in_directory(self, indir: str, extension=".csv", prefix=""):
        """
        Get list of files in a directory

        Parameters
        ----------
        indir : str
            input directory to search for files
        extension : str, optional
            extension of the files to search for
        prefix : str, optional
            string file name prefix to narrow search

        Returns
        -------
        list of str
            list of file path strings
        """
        raise NotImplementedError

    def clear_dir(self, dir_path: str):
        """
        Clear contents of existing directory

        Parameters
        ----------
        dir_path : str
            path to directory to be cleared
        """
        raise NotImplementedError

    def rm_dir(self, dir_path: str):
        """
        Delete existing directory

        Parameters
        ----------
        dir_path : str
            path to directory to be removed
        """
        raise NotImplementedError

    def rm_file(self, file_path: str):
        """
        Deletes existing file_path

        Parameters
        ----------
        file_path : str
            path to file to be removed
        """
        raise NotImplementedError



# Run from ml4ir/python directory as: python3 -m pytest ml4ir/applications/ranking/tests/test_create_dataset.py

import unittest
import os
import warnings
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.io.logging_utils import setup_logging
from ml4ir.applications.ranking.data.scripts.create_dataset import run_dataset_creation

warnings.filterwarnings("ignore")

ROOT_DATA_DIR = "ml4ir/applications/ranking/tests/data/csv/train"
FEATURE_CONFIG = "ml4ir/applications/ranking/tests/data/configs/feature_config.yaml"
OUTPUT_DIR = "ml4ir/applications/ranking/tests/test_output/synthetic"
LOG_DIR = "ml4ir/applications/ranking/tests/test_output_log/"


class RankingCreateDatasetTest(unittest.TestCase):
    def setUp(
        self,
        root_data_dir: str = ROOT_DATA_DIR,
        feature_config: str = FEATURE_CONFIG,
        output_dir: str = OUTPUT_DIR,
        log_dir: str = LOG_DIR,
    ):
        self.root_data_dir = root_data_dir
        self.feature_config = feature_config
        self.output_dir = output_dir
        self.log_dir = log_dir
        self.file_io = LocalIO()

        # Set up logging
        self.file_io.make_directory(self.log_dir, clear_dir=True)
        outfile: str = os.path.join(self.log_dir, "output_log.csv")
        self.logger = setup_logging(reset=True, file_name=outfile, log_to_file=True)

    def test_synthetic_data(self):

        feature_highval = {"text_match_bool": [0, 1]}
        max_num_records = 20
        num_samples = 10

        df = run_dataset_creation(
            self.root_data_dir,
            self.output_dir,
            self.feature_config,
            feature_highval,
            max_num_records,
            num_samples,
            random_state=123,
        )
        assert len(df) == 32
        assert df.query_id.nunique() == num_samples
        assert df.num_results_calc.max() <= max_num_records
        assert "text_match_bool" in list(df.columns)
        assert list(df.text_match_bool.unique()) == [0, 1]

        df_2 = run_dataset_creation(
            self.root_data_dir,
            self.output_dir,
            self.feature_config,
            feature_highval,
            max_num_records=2,
            num_samples=10,
            random_state=123,
        )
        assert len(df_2) == 20

    def tearDown(self):
        # Delete output directory
        self.file_io.rm_dir(self.output_dir)
        self.file_io.rm_dir(self.log_dir)


if __name__ == "__main__":
    unittest.main()



import os
import sys
from logging import Logger
from typing import Dict, Optional, List, Union, Type, Tuple
from tensorflow.keras import callbacks, Input, Model
from tensorflow.keras.optimizers import Optimizer
from tensorflow import data
from tensorflow.keras import metrics as kmetrics
import pandas as pd
import tensorflow as tf
import numpy as np

from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.io.file_io import FileIO
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.model.metrics.metrics_impl import get_metrics_impl
from ml4ir.base.model.scoring.scoring_model import ScorerBase, RelevanceScorer
from ml4ir.base.model.scoring.interaction_model import InteractionModel, UnivariateInteractionModel
from ml4ir.base.model.serving import define_serving_signatures
from ml4ir.base.model.scoring.prediction_helper import get_predict_fn
from ml4ir.base.model.callbacks.debugging import DebuggingCallback
from ml4ir.base.model.calibration.temperature_scaling import temperature_scale,\
    TemperatureScalingLayer
from ml4ir.applications.ranking.config.keys import PositionalBiasHandler
from ml4ir.base.config.keys import LearningRateScheduleKey


class RelevanceModelConstants:

    MODEL_PREDICTIONS_CSV_FILE = "model_predictions.csv"
    METRICS_CSV_FILE = "metrics.csv"
    GROUP_METRICS_CSV_FILE = "group_metrics.csv"
    CHECKPOINT_FNAME = "checkpoint.tf"


class RelevanceModel:
    def __init__(
        self,
        feature_config: FeatureConfig,
        tfrecord_type: str,
        file_io: FileIO,
        scorer: Optional[ScorerBase] = None,
        metrics: List[Union[Type[kmetrics.Metric], str]] = [],
        optimizer: Optional[Optimizer] = None,
        model_file: Optional[str] = None,
        initialize_layers_dict: dict = {},
        freeze_layers_list: list = [],
        compile_keras_model: bool = False,
        output_name: str = "score",
        logger=None,
    ):
        """
        Constructor to instantiate a RelevanceModel that can be used for
        training and evaluating the search ML task

        Parameters
        ----------
        feature_config : `FeatureConfig` object
            FeatureConfig object that defines the features to be loaded in the dataset
            and the preprocessing functions to be applied to each of them
        tfrecord_type : {"example", "sequence_example"}
            Type of the TFRecord protobuf message used for TFRecordDataset
        file_io : `FileIO` object
            file I/O handler objects for reading and writing data
        scorer : `ScorerBase` object
            Scorer object that wraps an InteractionModel and converts
            input features into scores
        metrics : list
            List of keras Metric classes that will be used for evaluating the trained model
        optimizer : `Optimizer`
            Tensorflow keras optimizer to be used for training the model
        model_file : str, optional
            Path to pretrained model file to be loaded for evaluation or retraining
        initialize_layers_dict : dict, optional
            Dictionary of tensorflow layer names mapped to the path of pretrained weights
            Use this for transfer learning with pretrained weights
        freeze_layers_list : list, optional
            List of model layer names to be frozen
            Use this for freezing pretrained weights from other ml4ir models
        compile_keras_model : bool, optional
            Whether the keras model loaded from disk should be compiled
            with loss, metrics and an optimizer
        output_name : str, optional
            Name of the output tensorflow node that captures the score
        logger : `Logger`, optional
            logging handler for status messages
        """
        self.feature_config: FeatureConfig = feature_config
        self.logger: Logger = logger
        self.output_name = output_name
        self.scorer = scorer
        self.tfrecord_type = tfrecord_type
        self.file_io = file_io

        if scorer:
            self.max_sequence_size = scorer.interaction_model.max_sequence_size
        else:
            self.max_sequence_size = 0

        # Load/Build Model
        if model_file and not compile_keras_model:
            """
            If a model file is specified, load it without compiling into a keras model

            NOTE:
            This will allow the model to be only used for inference and
            cannot be used for retraining.
            """
            self.model: Model = self.load(model_file)
            self.is_compiled = False
        else:

            """
            Specify inputs to the model

            Individual input nodes are defined for each feature
            Each data point represents features for all records in a single query
            """
            inputs: Dict[str, Input] = feature_config.define_inputs()
            scores, train_features, metadata_features = scorer(inputs)

            # Create model with functional Keras API
            self.model = Model(inputs=inputs, outputs={self.output_name: scores})
            self.model.output_names = [self.output_name]

            # Get loss fn
            loss_fn = scorer.loss.get_loss_fn(**metadata_features)

            # Get metric objects
            metrics_impl: List[Union[str, kmetrics.Metric]] = get_metrics_impl(
                metrics=metrics, feature_config=feature_config, metadata_features=metadata_features
            )

            # Compile model
            """
            NOTE:
            Related Github issue: https://github.com/tensorflow/probability/issues/519
            """
            self.model.compile(
                optimizer=optimizer,
                loss=loss_fn,
                metrics=metrics_impl,
                experimental_run_tf_function=False,
            )

            # Write model summary to logs
            model_summary = list()
            self.model.summary(print_fn=lambda x: model_summary.append(x))
            if self.logger:
                self.logger.info("\n".join(model_summary))

            if model_file:
                """
                If model file is specified, load the weights from the SavedModel

                NOTE:
                The architecture, loss and metrics of self.model need to
                be the same as the loaded SavedModel
                """
                self.load_weights(model_file)

            # Initialize layer weights
            for layer_name, layer_file in initialize_layers_dict.items():
                layer = self.model.get_layer(layer_name)
                layer.set_weights(self.file_io.load_numpy_array(layer_file, unzip=True))
                self.logger.info("Setting {} weights from {}".format(layer_name, layer_file))

            # Freeze layer weights
            for layer_name in freeze_layers_list:
                layer = self.model.get_layer(layer_name)
                layer.trainable = False
                self.logger.info("Freezing {} layer".format(layer_name))

            self.is_compiled = True

    @classmethod
    def from_relevance_scorer(
        cls,
        feature_config: FeatureConfig,
        interaction_model: InteractionModel,
        model_config: dict,
        loss: RelevanceLossBase,
        metrics: List[Union[kmetrics.Metric, str]],
        optimizer: Optimizer,
        tfrecord_type: str,
        file_io: FileIO,
        model_file: Optional[str] = None,
        initialize_layers_dict: dict = {},
        freeze_layers_list: list = [],
        compile_keras_model: bool = False,
        output_name: str = "score",
        logger=None,
    ):
        """
        Create a RelevanceModel with default Scorer function
        constructed from an InteractionModel

        Parameters
        ----------
        feature_config : `FeatureConfig` object
            FeatureConfig object that defines the features to be loaded in the dataset
            and the preprocessing functions to be applied to each of them
        tfrecord_type : {"example", "sequence_example"}
            Type of the TFRecord protobuf message used for TFRecordDataset
        file_io : `FileIO` object
            file I/O handler objects for reading and writing data
        interaction_model : `InteractionModel` object
            InteractionModel object that converts input features into a
            dense feature representation
        loss : `RelevanceLossBase` object
            Loss object defining the final activation layer and the loss function
        metrics : list
            List of keras Metric classes that will be used for evaluating the trained model
        optimizer : `Optimizer`
            Tensorflow keras optimizer to be used for training the model
        model_file : str, optional
            Path to pretrained model file to be loaded for evaluation or retraining
        initialize_layers_dict : dict, optional
            Dictionary of tensorflow layer names mapped to the path of pretrained weights
            Use this for transfer learning with pretrained weights
        freeze_layers_list : list, optional
            List of model layer names to be frozen
            Use this for freezing pretrained weights from other ml4ir models
        compile_keras_model : bool, optional
            Whether the keras model loaded from disk should be compiled
            with loss, metrics and an optimizer
        output_name : str, optional
            Name of the output tensorflow node that captures the score
        logger : `Logger`, optional
            logging handler for status messages

        Returns
        -------
        RelevanceModel
            RelevanceModel object with a default scorer build with a custom
            InteractionModel
        """
        assert isinstance(interaction_model, InteractionModel)
        assert isinstance(loss, RelevanceLossBase)

        scorer: ScorerBase = RelevanceScorer(
            model_config=model_config,
            interaction_model=interaction_model,
            loss=loss,
            output_name=output_name,
        )

        return cls(
            scorer=scorer,
            feature_config=feature_config,
            metrics=metrics,
            optimizer=optimizer,
            tfrecord_type=tfrecord_type,
            model_file=model_file,
            initialize_layers_dict=initialize_layers_dict,
            freeze_layers_list=freeze_layers_list,
            compile_keras_model=compile_keras_model,
            output_name=output_name,
            file_io=file_io,
            logger=logger,
        )

    @classmethod
    def from_univariate_interaction_model(
        cls,
        model_config,
        feature_config: FeatureConfig,
        tfrecord_type: str,
        loss: RelevanceLossBase,
        metrics: List[Union[kmetrics.Metric, str]],
        optimizer: Optimizer,
        feature_layer_keys_to_fns: dict = {},
        model_file: Optional[str] = None,
        initialize_layers_dict: dict = {},
        freeze_layers_list: list = [],
        compile_keras_model: bool = False,
        output_name: str = "score",
        max_sequence_size: int = 0,
        file_io: FileIO = None,
        logger=None,
    ):
        """
        Create a RelevanceModel with default UnivariateInteractionModel

        Parameters
        ----------
        feature_config : `FeatureConfig` object
            FeatureConfig object that defines the features to be loaded in the dataset
            and the preprocessing functions to be applied to each of them
        model_config : dict
            dictionary defining the dense model architecture
        tfrecord_type : {"example", "sequence_example"}
            Type of the TFRecord protobuf message used for TFRecordDataset
        file_io : `FileIO` object
            file I/O handler objects for reading and writing data
        loss : `RelevanceLossBase` object
            Loss object defining the final activation layer and the loss function
        metrics : list
            List of keras Metric classes that will be used for evaluating the trained model
        optimizer : `Optimizer`
            Tensorflow keras optimizer to be used for training the model
        feature_layer_keys_to_fns : dict
            Dictionary of custom feature transformation functions to be applied
            on the input features as part of the InteractionModel
        model_file : str, optional
            Path to pretrained model file to be loaded for evaluation or retraining
        initialize_layers_dict : dict, optional
            Dictionary of tensorflow layer names mapped to the path of pretrained weights
            Use this for transfer learning with pretrained weights
        freeze_layers_list : list, optional
            List of model layer names to be frozen
            Use this for freezing pretrained weights from other ml4ir models
        compile_keras_model : bool, optional
            Whether the keras model loaded from disk should be compiled
            with loss, metrics and an optimizer
        output_name : str, optional
            Name of the output tensorflow node that captures the score
        max_sequence_size : int, optional
            Maximum length of the sequence to be used for SequenceExample protobuf objects
        logger : `Logger`, optional
            logging handler for status messages

        Returns
        -------
        RelevanceModel
            RelevanceModel object with a UnivariateInteractionModel
        """

        interaction_model: InteractionModel = UnivariateInteractionModel(
            feature_config=feature_config,
            feature_layer_keys_to_fns=feature_layer_keys_to_fns,
            tfrecord_type=tfrecord_type,
            max_sequence_size=max_sequence_size,
        )

        return cls.from_relevance_scorer(
            interaction_model=interaction_model,
            model_config=model_config,
            feature_config=feature_config,
            loss=loss,
            metrics=metrics,
            optimizer=optimizer,
            tfrecord_type=tfrecord_type,
            model_file=model_file,
            initialize_layers_dict=initialize_layers_dict,
            freeze_layers_list=freeze_layers_list,
            compile_keras_model=compile_keras_model,
            output_name=output_name,
            file_io=file_io,
            logger=logger,
        )

    def define_scheduler_as_callback(self, monitor_metric, model_config):
        """
        Adding reduce lr on plateau as a callback if specified

        Parameters
        ----------
        monitor_metric : string
           The metric to be monitored by the callback
        model_config : dict
            dictionary defining the dense model architecture

        Returns
        -------
        reduce_lr
            The created scheduler callback object.
        """

        if model_config and 'lr_schedule' in model_config:
            lr_schedule = model_config['lr_schedule']
            lr_schedule_key = lr_schedule['key']
            if lr_schedule_key == LearningRateScheduleKey.REDUCE_LR_ON_PLATEAU:
                if monitor_metric is None:
                    reduce_lr = tf.keras.callbacks.ReduceLROnPlateau(factor=lr_schedule.get('factor', 0.5),
                                                                     patience=lr_schedule.get('patience', 5),
                                                                     min_lr=lr_schedule.get('min_lr', 0.0001),
                                                                     mode=lr_schedule.get('mode', 'auto'),
                                                                     verbose=1)
                else:
                    if not monitor_metric.startswith("val_"):
                        monitor_metric = "val_{}".format(monitor_metric)
                    reduce_lr = tf.keras.callbacks.ReduceLROnPlateau(monitor=monitor_metric,
                                                                     factor=lr_schedule.get('factor', 0.5),
                                                                     patience=lr_schedule.get('patience', 5),
                                                                     min_lr=lr_schedule.get('min_lr', 0.0001),
                                                                     mode=lr_schedule.get('mode', 'auto'),
                                                                     verbose=1)
                return reduce_lr

    def fit(
        self,
        dataset: RelevanceDataset,
        num_epochs: int,
        models_dir: str,
        logs_dir: Optional[str] = None,
        logging_frequency: int = 25,
        monitor_metric: str = "",
        monitor_mode: str = "",
        patience=2,
    ):
        """
        Trains model for defined number of epochs
        and returns the training and validation metrics as a dictionary

        Parameters
        ----------
        dataset : `RelevanceDataset` object
            RelevanceDataset object to be used for training and validation
        num_epochs : int
            Value specifying number of epochs to train for
        models_dir : str
            Directory to save model checkpoints
        logs_dir : str, optional
            Directory to save model logs
            If set to False, no progress logs will be written
        logging_frequency : int, optional
            Every #batches to log results
        monitor_metric : str, optional
            Name of the metric to monitor for early stopping, checkpointing
        monitor_mode : {"max", "min"}
            Whether to maximize or minimize the monitoring metric
        patience : int
            Number of epochs to wait before early stopping

        Returns
        -------
        train_metrics : dict
            Train and validation metrics in a single dictionary
            where key is metric name and value is floating point metric value.
            This dictionary will be used for experiment tracking for each ml4ir run
        """

        if not monitor_metric.startswith("val_"):
            monitor_metric = "val_{}".format(monitor_metric)
        callbacks_list: list = self._build_callback_hooks(
            models_dir=models_dir,
            logs_dir=logs_dir,
            is_training=True,
            logging_frequency=logging_frequency,
            monitor_mode=monitor_mode,
            monitor_metric=monitor_metric,
            patience=patience,
        )

        if self.is_compiled:
            history = self.model.fit(
                x=dataset.train,
                validation_data=dataset.validation,
                epochs=num_epochs,
                verbose=True,
                callbacks=callbacks_list,
            )

            # Write metrics for experiment tracking
            # Returns a dictionary
            train_metrics = dict()
            for metric, value in history.history.items():
                if not metric.startswith("val_"):
                    """
                    NOTE:
                    Prepend "train_" to metrics on training dataset
                    to differentiate from validation and test metrics
                    in the final experiment results
                    """
                    # History is a dict of key: list(values per epoch)
                    # We are capturing the metrics of the last epoch (-1)
                    train_metrics["train_{}".format(metric)] = value[-1]
                else:
                    train_metrics[metric] = value[-1]

            return train_metrics
        else:
            raise NotImplementedError(
                "The model could not be trained. "
                "Check if the model was compiled correctly."
                " Training loaded SavedModel is not currently supported."
            )

    def predict(
        self,
        test_dataset: data.TFRecordDataset,
        inference_signature: str = "serving_default",
        additional_features: dict = {},
        logs_dir: Optional[str] = None,
        logging_frequency: int = 25,
    ):
        """
        Predict the scores on the test dataset using the trained model

        Parameters
        ----------
        test_dataset : `Dataset` object
            `Dataset` object for which predictions are to be made
        inference_signature : str, optional
            If using a SavedModel for prediction, specify the inference signature to be used for computing scores
        additional_features : dict, optional
            Dictionary containing new feature name and function definition to
            compute them. Use this to compute additional features from the scores.
            For example, converting ranking scores for each document into ranks for
            the query
        logs_dir : str, optional
            Path to directory to save logs
        logging_frequency : int
            Value representing how often(in batches) to log status

        Returns
        -------
        `pd.DataFrame`
            pandas DataFrame containing the predictions on the test dataset
            made with the `RelevanceModel`
        """
        if logs_dir:
            outfile = os.path.join(logs_dir, RelevanceModelConstants.MODEL_PREDICTIONS_CSV_FILE)
            # Delete file if it exists
            self.file_io.rm_file(outfile)

        _predict_fn = get_predict_fn(
            model=self.model,
            tfrecord_type=self.tfrecord_type,
            feature_config=self.feature_config,
            inference_signature=inference_signature,
            is_compiled=self.is_compiled,
            output_name=self.output_name,
            features_to_return=self.feature_config.get_features_to_log(),
            additional_features=additional_features,
            max_sequence_size=self.max_sequence_size,
        )

        predictions_df_list = list()
        batch_count = 0
        for predictions_dict in test_dataset.map(_predict_fn).take(-1):
            predictions_df = pd.DataFrame(predictions_dict)

            if logs_dir:
                np.set_printoptions(
                    formatter={"all": lambda x: str(x.decode("utf-8"))
                               if isinstance(x, bytes) else str(x)},
                    linewidth=sys.maxsize, threshold=sys.maxsize)  # write the full line in the csv not the truncated version.

                # Decode bytes features to strings
                for col in predictions_df.columns:
                    if isinstance(predictions_df[col].values[0], bytes):
                        predictions_df[col] = predictions_df[col].str.decode("utf8")

                if os.path.isfile(outfile):
                    predictions_df.to_csv(outfile, mode="a", header=False, index=False)
                else:
                    # If writing first time, write headers to CSV file
                    predictions_df.to_csv(outfile, mode="w", header=True, index=False)

            else:
                predictions_df_list.append(predictions_df)

            batch_count += 1
            if batch_count % logging_frequency == 0:
                self.logger.info("Finished predicting scores for {} batches".format(batch_count))

        predictions_df = None
        if logs_dir:
            self.logger.info("Model predictions written to -> {}".format(outfile))
        else:
            predictions_df = pd.concat(predictions_df_list)

        return predictions_df

    def evaluate(
        self,
        test_dataset: data.TFRecordDataset,
        inference_signature: str = None,
        additional_features: dict = {},
        group_metrics_min_queries: int = 50,
        logs_dir: Optional[str] = None,
        logging_frequency: int = 25,
        compute_intermediate_stats: bool = True,
    ):
        """
        Evaluate the RelevanceModel

        Parameters
        ----------
        test_dataset: an instance of tf.data.dataset
        inference_signature : str, optional
            If using a SavedModel for prediction, specify the inference signature to be used for computing scores
        additional_features : dict, optional
            Dictionary containing new feature name and function definition to
            compute them. Use this to compute additional features from the scores.
            For example, converting ranking scores for each document into ranks for
            the query
        group_metrics_min_queries : int, optional
            Minimum count threshold per group to be considered for computing
            groupwise metrics
        logs_dir : str, optional
            Path to directory to save logs
        logging_frequency : int
            Value representing how often(in batches) to log status
        compute_intermediate_stats : bool
            Determines if group metrics and other intermediate stats on the test set should be computed

        Returns
        -------
        df_overall_metrics : `pd.DataFrame` object
            `pd.DataFrame` containing overall metrics
        df_groupwise_metrics : `pd.DataFrame` object
            `pd.DataFrame` containing groupwise metrics if
            group_metric_keys are defined in the FeatureConfig
        metrics_dict : dict
            metrics as a dictionary of metric names mapping to values

        Notes
        -----
        You can directly do a `model.evaluate()` only if the keras model is compiled

        Override this method to implement your own evaluation metrics.
        """
        if self.is_compiled:
            metrics_dict = self.model.evaluate(test_dataset)
            return None, None, dict(zip(self.model.metrics_names, metrics_dict))
        else:
            raise NotImplementedError

    def run_ttest(self, mean, variance, n, ttest_pvalue_threshold):
        """
        Compute the paired t-test statistic and its p-value given mean, standard deviation and sample count
        Parameters
        ----------
        mean: float
            The mean of the rank differences for the entire dataset
        variance: float
            The variance of the rank differences for the entire dataset
        n: int
            The number of samples in the entire dataset
        ttest_pvalue_threshold: float
            P-value threshold for student t-test
        metrics_dict: dict
            dictionary of metrics to keep track

        Returns
        -------
        t_test_metrics_dict: Dictionary
            A dictionary with the t-test metrics recorded.
        """
        raise NotImplementedError

    def save(
        self,
        models_dir: str,
        preprocessing_keys_to_fns={},
        postprocessing_fn=None,
        required_fields_only: bool = True,
        pad_sequence: bool = False,
        sub_dir: str = "final",
        dataset: Optional[RelevanceDataset] = None,
        experiment_details: Optional[dict] = None
    ):
        """
        Save the RelevanceModel as a tensorflow SavedModel to the `models_dir`

        There are two different serving signatures currently used to save the model:

        * `default`: default keras model without any pre/post processing wrapper

        * `tfrecord`: serving signature that allows keras model to be served using TFRecord proto messages.
                  Allows definition of custom pre/post processing logic

        Additionally, each model layer is also saved as a separate numpy zipped
        array to enable transfer learning with other ml4ir models.

        Parameters
        ----------
        models_dir : str
            path to directory to save the model
        preprocessing_keys_to_fns : dict
            dictionary mapping function names to tf.functions that should be
            saved in the preprocessing step of the tfrecord serving signature
        postprocessing_fn: function
            custom tensorflow compatible postprocessing function to be used at serving time.
            Saved as part of the postprocessing layer of the tfrecord serving signature
        required_fields_only: bool
            boolean value defining if only required fields
            need to be added to the tfrecord parsing function at serving time
        pad_sequence: bool, optional
            Value defining if sequences should be padded for SequenceExample proto inputs at serving time.
            Set this to False if you want to not handle padded scores.
        sub_dir: str, optional
            sub directory name to save the model into
        dataset : `RelevanceDataset` object
            RelevanceDataset object that can optionally be passed to be used by downstream jobs
            that want to save the data along with the model.
            Note that this feature is currently unimplemented and is upto the users to override
            and customize.
        experiment_details: dict
            Dictionary containing metadata and results about the current experiment

        Notes
        -----
        All the functions passed under `preprocessing_keys_to_fns` here must be
        serializable tensor graph operations
        """

        model_file = os.path.join(models_dir, sub_dir)

        # Save model with default signature
        self.model.save(filepath=os.path.join(model_file, "default"))

        """
        Save model with custom signatures

        Currently supported
        - signature to read TFRecord SequenceExample inputs
        """
        self.model.save(
            filepath=os.path.join(model_file, "tfrecord"),
            signatures=define_serving_signatures(
                model=self.model,
                tfrecord_type=self.tfrecord_type,
                feature_config=self.feature_config,
                preprocessing_keys_to_fns=preprocessing_keys_to_fns,
                postprocessing_fn=postprocessing_fn,
                required_fields_only=required_fields_only,
                pad_sequence=pad_sequence,
                max_sequence_size=self.max_sequence_size,
            ),
        )

        # Save individual layer weights
        self.file_io.make_directory(os.path.join(model_file, "layers"), clear_dir=True)
        for layer in self.model.layers:
            try:
                self.file_io.save_numpy_array(
                    np_array=layer.get_weights(),
                    file_path=os.path.join(model_file, "layers", "{}.npz".format(layer.name)),
                    zip=True,
                )
            except FileNotFoundError:
                self.logger.warning(
                    "Error saving layer: {} due to FileNotFoundError. Skipping...".format(layer.name))

        self.logger.info("Final model saved to : {}".format(model_file))

    def load(self, model_file: str) -> Model:
        """
        Loads model from the SavedModel file specified

        Parameters
        ----------
        model_file : str
            path to file with saved tf keras model

        Returns
        -------
        `tf.keras.Model`
            Tensorflow keras model loaded from file

        Notes
        -----
        Retraining currently not supported!
        Would require compiling the model with the right loss and optimizer states
        """
        """
        NOTE:
        There is currently a bug in Keras Model with saving/loading
        models with custom losses and metrics.

        Therefore, we are currently loading the SavedModel with compile=False
        The saved model signatures can be used for inference at serving time

        Ref:
        https://github.com/keras-team/keras/issues/5916
        https://github.com/tensorflow/tensorflow/issues/32348
        https://github.com/keras-team/keras/issues/3977

        """
        model = tf.keras.models.load_model(model_file, compile=False)

        self.logger.info("Successfully loaded SavedModel from {}".format(model_file))
        self.logger.warning("Retraining is not yet supported. Model is loaded with compile=False")

        return model

    def load_weights(self, model_file: str):
        """
        Load saved model with compile=False

        Parameters
        ----------
        model_file : str
            path to file with saved tf keras model
        """
        loaded_model = self.load(model_file)

        # Set weights of Keras model from the loaded model weights
        self.model.set_weights(loaded_model.get_weights())
        self.logger.info("Weights have been set from SavedModel. RankingModel can now be trained.")

    def _build_callback_hooks(
        self,
        models_dir: str,
        logs_dir: Optional[str] = None,
        is_training=True,
        logging_frequency=25,
        monitor_metric: str = "",
        monitor_mode: str = "",
        patience=2,
    ):
        """
        Build callback hooks for the training and evaluation loop

        Parameters
        ----------
        models_dir : str
            Path to directory to save model checkpoints
        logs_dir : str
            Path to directory to save tensorboard logs
        is_training : bool, optional
            Whether we are building callbacks for training or evaluation
        logging_frequency : int, optional
            How often, in number of epochs, to log training and evaluation progress
        monitor_metric : str, optional
            Name of metric to be used for ModelCheckpoint and EarlyStopping callbacks
        monitor_mode : {"max", "min"}, optional
            Mode for maximizing or minimizing the ModelCheckpoint and EarlyStopping
        patience : int, optional
            Number of epochs to wait before early stopping if metric change is below tolerance

        Returns
        -------
        callbacks_list : list
            List of callbacks to be used with the RelevanceModel training and evaluation
        """
        callbacks_list: list = list()

        if is_training:
            # Model checkpoint
            if models_dir and monitor_metric:
                checkpoints_path = os.path.join(
                    models_dir, RelevanceModelConstants.CHECKPOINT_FNAME
                )
                cp_callback = callbacks.ModelCheckpoint(
                    filepath=checkpoints_path,
                    save_weights_only=False,
                    verbose=1,
                    save_best_only=True,
                    mode=monitor_mode,
                    monitor=monitor_metric,
                )
                callbacks_list.append(cp_callback)

            # Early Stopping
            if monitor_metric:
                early_stopping_callback = callbacks.EarlyStopping(
                    monitor=monitor_metric,
                    mode=monitor_mode,
                    patience=patience,
                    verbose=1,
                    restore_best_weights=True,
                )
                callbacks_list.append(early_stopping_callback)

        # TensorBoard
        if logs_dir:
            tensorboard_callback = callbacks.TensorBoard(
                log_dir=logs_dir, histogram_freq=1, update_freq=5
            )
            callbacks_list.append(tensorboard_callback)

        # Debugging/Logging
        callbacks_list.append(DebuggingCallback(self.logger, logging_frequency))

        # Adding lr scheduler as a callback; used for `ReduceLROnPlateau` which we treat today as a callback
        scheduler_callback = self.define_scheduler_as_callback(monitor_metric, self.scorer.model_config)
        if scheduler_callback:
            callbacks_list.append(scheduler_callback)

        # Add more here

        return callbacks_list

    def calibrate(self, relevance_dataset, logger, logs_dir_local, **kwargs)\
            -> Tuple[np.ndarray, ...]:
        """Calibrate model with temperature scaling
        Parameters
        ----------
        relevance_dataset: RelevanceDataset
            RelevanceDataset object to be used for training and evaluating temperature scaling
        logger: Logger
            Logger object to log events
        logs_dir_local: str
            path to save the calibration results. (zipped csv file containing original
            probabilities, calibrated probabilities, ...)
        Returns
        -------
        `Union[np.ndarray, Tuple[np.ndarray, ...]]`
        optimizer output containing temperature value learned during temperature scaling

        """
        logger.info("=" * 50)
        logger.info("Calibrating the model with temperature scaling")
        return temperature_scale(model=self.model,
                                 scorer=self.scorer,
                                 dataset=relevance_dataset,
                                 logger=logger,
                                 logs_dir_local=logs_dir_local,
                                 file_io=self.file_io,
                                 **kwargs)

    def add_temperature_layer(self, temperature: float = 1.0, layer_name: str = 'temperature_layer'):
        """Add temperature layer to the input of last activation (softmax) layer
        Parameters
        ----------
        self: RelevanceModel
                        input RelevanceModel object that its last layer inputs will be divided by a
                        temperature value
        temperature: float
                     a scalar value to scale the last activation layer inputs
        layer_name: str
            name of the temperature scaling layer
        Returns
        -------
        `RelevanceModel`
        updated RelevanceModel object with temperature
        """

        # get  last layer's output  --> MUST **NOT** BE AN ACTIVATION (e.g. SOFTMAX) LAYER
        final_layer_name = self.scorer.model_config['layers'][-1]['name']

        final_layer = self.model.get_layer(name=final_layer_name).output
        temperature_layer = TemperatureScalingLayer(name=layer_name,
                                                    temperature=temperature)(final_layer)

        # using the `last layer` as final activation function before computing loss
        idx_activation = -1
        if len(self.model.layers) > 0 and isinstance(self.model.layers[idx_activation],
                                                     tf.keras.layers.Activation):
            # creating new activation layer
            activation_layer_name = self.model.get_layer(index=idx_activation).name
            activation_function = self.model.get_layer(index=idx_activation).activation
            activation_layer = tf.keras.layers.Activation(
                activation_function, name=activation_layer_name)(temperature_layer)
            # creating new keras Functional API model
            self.model = Model(self.model.inputs, activation_layer)
            self.logger.info(f'Temperature Scaling layer added and new Functional API model'
                             f' replaced; temperature = {temperature}.')
        else:
            self.logger.info("Skipping adding Temperature Scaling layer because no activation "
                             "exist in the last layer of Keras original model!")



from ml4ir.base.model.losses.loss_base import RelevanceLossBase
import ml4ir.applications.ranking.model.losses.pointwise_losses as pointwise_losses
import ml4ir.applications.ranking.model.losses.pairwise_losses as pairwise_losses
import ml4ir.applications.ranking.model.losses.listwise_losses as listwise_losses
from ml4ir.applications.ranking.config.keys import LossKey


def get_loss(loss_key, scoring_type) -> RelevanceLossBase:
    """
    Factor method to get Loss function object

    Parameters
    ----------
    loss_key : str
        Name of the loss function as specified by LossKey
    scoring_type : str
        Type of scoring function - pointwise, pairwise, groupwise

    Returns
    -------
    RelevanceLossBase
        RelevanceLossBase object that applies the final activation layer
        and computes the loss function from the model score
    """
    if loss_key == LossKey.SIGMOID_CROSS_ENTROPY:
        return pointwise_losses.SigmoidCrossEntropy(loss_key=loss_key, scoring_type=scoring_type)
    elif loss_key == LossKey.RANK_ONE_LISTNET:
        return listwise_losses.RankOneListNet(loss_key=loss_key, scoring_type=scoring_type)
    elif loss_key == LossKey.SOFTMAX_CROSS_ENTROPY:
        return listwise_losses.SoftmaxCrossEntropy(loss_key=loss_key, scoring_type=scoring_type)
    else:
        raise NotImplementedError



import os
import numpy as np
import tensorflow as tf
from tensorflow.keras import models as kmodels

from ml4ir.applications.classification.pipeline import ClassificationPipeline
from ml4ir.applications.classification.tests.test_base import ClassificationTestBase
from ml4ir.base.config.keys import ServingSignatureKey
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.features.preprocessing import split_and_pad_string
from ml4ir.base.model.relevance_model import RelevanceModel


class ClassificationServingTest(ClassificationTestBase):
    """Assess model serving."""

    def test_serve_default_pipeline(self):
        """
        Train a simple model and test serving flow by loading the SavedModel
        """
        # Test model training on TFRecord Example data
        self.set_seeds()
        classification_pipeline: ClassificationPipeline = ClassificationPipeline(
            args=self.get_overridden_args()
        )

        parsed_relevance_dataset: RelevanceDataset = classification_pipeline.get_relevance_dataset()
        raw_relevance_dataset: RelevanceDataset = classification_pipeline.get_relevance_dataset(
            parse_tfrecord=False
        )
        classification_model: RelevanceModel = classification_pipeline.get_relevance_model()

        classification_model.fit(
            dataset=parsed_relevance_dataset, num_epochs=1, models_dir=self.output_dir
        )

        preprocessing_keys_to_fns = {"split_and_pad_string": split_and_pad_string}

        classification_model.save(
            models_dir=self.args.models_dir,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            required_fields_only=True,
        )

        default_model = kmodels.load_model(
            os.path.join(self.output_dir, "final", "default"), compile=False
        )
        assert ServingSignatureKey.DEFAULT in default_model.signatures
        default_signature = default_model.signatures[ServingSignatureKey.DEFAULT]

        tfrecord_model = kmodels.load_model(
            os.path.join(self.output_dir, "final", "tfrecord"), compile=False
        )
        assert ServingSignatureKey.TFRECORD in tfrecord_model.signatures
        tfrecord_signature = tfrecord_model.signatures[ServingSignatureKey.TFRECORD]

        # Fetch a single batch for testing
        sequence_example_protos = next(iter(raw_relevance_dataset.test))
        parsed_sequence_examples = next(iter(parsed_relevance_dataset.test))[0]
        parsed_dataset_batch = parsed_relevance_dataset.test.take(1)

        # Use the loaded serving signatures for inference
        model_predictions = classification_model.predict(parsed_dataset_batch)[
            self.args.output_name
        ].values
        default_signature_predictions = default_signature(**parsed_sequence_examples)[
            self.args.output_name
        ]

        # Since we do not pad dummy records in tfrecord serving signature,
        # we can only predict on a single record at a time
        tfrecord_signature_predictions = [
            tfrecord_signature(protos=tf.gather(sequence_example_protos, [i]))[
                self.args.output_name
            ]
            for i in range(self.args.batch_size)
        ]

        # Compare the scores from the different versions of the model
        assert np.isclose(model_predictions[0], default_signature_predictions[0], rtol=0.01,).all()
        assert np.isclose(
            model_predictions[0], tfrecord_signature_predictions[0], rtol=0.01,
        ).all()



"""Implementing temperature scaling proposed in https://github.com/gpleiss/temperature_scaling
The tensorflow implementation and functions are inspired by
https://www.tensorflow.org/probability/examples/Optimizers_in_TensorFlow_Probability
"""
import sys
import os
import contextlib
import functools
import time
import shutil

from logging import Logger
from collections import Callable
from typing import Union, Tuple

import tensorflow as tf
import numpy as np
import pandas as pd
import tensorflow_probability as tfp

from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.model.scoring.scoring_model import ScorerBase
from ml4ir.base.io.file_io import FileIO

TEMPERATURE_SCALE = 'temp_scaling_scores'


def make_val_and_grad_fn(fun: Callable) -> Callable:
    """Returns a function that computes y = fn(*args, **kwargs) and gradients of y wrt
    each of args and kwargs
    Parameters
    ---------
        fun : Callable
             Python callable to be differentiated
    """
    @functools.wraps(fun)
    def val_and_grad(variables):
        return tfp.math.value_and_gradient(fun, variables)
    return val_and_grad


@contextlib.contextmanager
def timed_execution(logger: Logger) -> None:
    """A context manager to compute the execution time of a process usually within
     a 'with' statement
    Parameter
    --------
        logger: logging.Logger object
                Logger object used for logging

    """
    initial_time = time.time()
    yield
    duration = time.time() - initial_time
    logger.info(f'Evaluation took: {duration} seconds')


def np_value(tensor: Union[tf.Tensor, Tuple[tf.Tensor, ...]]) -> Union[np.ndarray,
                                                                       Tuple[np.ndarray, ...]]:
    """Computes numpy value out of possibly nested tuple of tensors.
    Parameters
    ---------
        tensor: Union[tf.Tensor, Tuple[tf.Tensor, ...]]
                a tensor or a tuple of tensors object
    Returns
    ------
        `Union[np.ndarray, Tuple[np.ndarray, ...]`
         numpy value out of input tensor
    """
    if isinstance(tensor, tuple):
        return type(tensor)(*(np_value(t) for t in tensor))
    return tensor.numpy()


def run_optimizer(optimizer: Callable, logger: Logger) -> Tuple[np.ndarray, ...]:
    """Runs an optimizer and measure it's execution time.
    Parameters
    ---------
        optimizer: function
            It is used for optimization
        logger: Logger
            Logger object used for logging
    Returns
    ------
        `np.ndarray`
        output of the optimizer function
    """
    optimizer()  # Warmup.
    with timed_execution(logger):
        result = optimizer()
    return np_value(result)


def accuracy(scores, labels) -> np.ndarray:
    """Computes accuracy.
    Parameters
    ---------
        scores: tf.Tensor
                tf.Tensor object that contains softmaxes
        labels: tf.Tensor
                tf.Tensor object that contains labels
    Returns
    -------
    `np.ndarray`
    accuracy value

    """
    acc = tf.keras.metrics.Accuracy()
    acc.update_state(labels, tf.argmax(scores, axis=-1))
    return acc.result().numpy()


def dict_to_csv(data_dic: dict, root_dir: str, file_io: FileIO, zip_output: bool = True) -> str:
    """Saves input dictionary to a csv file and zips if requested
    Parameters
    ---------
        data_dic: dict
                input dict to be converted to a zipped csv file
        root_dir: str
                path to save the output file
        file_io: FileIO
                file I/O handler objects for reading and writing data
        zip_output: bool
                boolean value indicates whether the output should be zipped
    Returns
    -------
        `str`
         path to the created zip file
    """
    # creating zip dir
    final_dir_path = os.path.join(root_dir, TEMPERATURE_SCALE)
    file_io.make_directory(final_dir_path)

    csv_path = os.path.join(final_dir_path, f'{TEMPERATURE_SCALE}.csv')

    # creating .csv
    pd.DataFrame.from_dict(data_dic).to_csv(csv_path, index=False)
    if zip_output:
        # creating .zip file
        shutil.make_archive(final_dir_path, "zip", root_dir, TEMPERATURE_SCALE)

        # removing the dir, keeping only the zip
        shutil.rmtree(final_dir_path)
    return final_dir_path


def get_intermediate_model(model, scorer) -> tf.keras.models.Model:
    """Creates a tf.keras.models.Model copy of `model`. This intermediate model must generate
    logits (inputs of softmax).

    Parameters
    ----------
        model: tf.keras.models.Model
                Model object to get the intermediate model from
        scorer: ScorerBase
                scorerBase object to get the `model_config` from

    Returns
    ------
    tf.keras.models.Model
    a tf.keras.models.Model copy of the `model`
    """
    # get  last layer's output  --> MUST **NOT** BE AN ACTIVATION (e.g. SOFTMAX) LAYER
    final_layer_name = scorer.model_config['layers'][-1]['name']
    layer_output = model.get_layer(name=final_layer_name).output

    return tf.keras.models.Model(inputs=model.input, outputs=layer_output)


def eval_relevance_model(scorer: ScorerBase, logits: np.ndarray, labels, temperature=None):
    """Evaluates the relevance model given the logits and labels
    Parameters
    ----------
        scorer: ScorerBase
                Scorer of the RelevanceModel
        logits: numpy.ndarray
                input of softmax
        labels: numpy.ndarray
                class labels
        temperature: TF.Tensor
                temperature parameter of size=(1)
    Returns
    -------
        `np.ndarray`
         accuracy value,
         `tf.Tenosr`
         NLL loss,
         `tf.Tensor`
         softmaxes
    """
    # convert to TF.Tensor
    logits_tensor = tf.constant(logits, name='logits_test', dtype=tf.float32)
    labels_tensor = tf.constant(labels, name='labels_test', dtype=tf.int32)

    if temperature:
        logits_tensor = tf.divide(logits_tensor, temperature)
    nll = tf.reduce_mean(
        tf.nn.sparse_softmax_cross_entropy_with_logits(labels=labels_tensor, logits=logits_tensor))

    # getting the softmaxes using the model final activation function
    softmaxes = scorer.final_activation_op(logits_tensor, {})
    return accuracy(softmaxes, labels_tensor), tf.identity(nll), softmaxes


def get_logits_labels(model: tf.keras.Model, evaluation_set: tf.data.TFRecordDataset):
    """Predicts model output on the given evaluation set
    Parameters:
        model: tf.keras.Model to be used for prediction
        evaluation_set: tf.data.TFRecordDataset to be used for prediction
    Returns:
        model output: numpy.ndarray
        labels: numpy.ndarray
    """
    logits_numpys = model.predict(evaluation_set).squeeze()
    labels_numpys = np.concatenate(list(map(lambda x: x[1], evaluation_set))).squeeze().argmax(
        axis=-1)
    return logits_numpys, labels_numpys


def temperature_scale(model: tf.keras.Model,
                      scorer: ScorerBase,
                      dataset: RelevanceDataset,
                      logger: Logger,
                      logs_dir_local: str,
                      temperature: float,
                      file_io: FileIO,
                      zip_output: bool = True
                      ) -> Tuple[np.ndarray, ...]:
    """learns a temperature parameter using Temperature Scaling (TS) technique on the validation set
    It, then, computes the probability scores of the test set with and without TS and writes them
    in a .zip file.

    Parameters
    ----------
        model :  tf.keras.Model
                 Model object to be used for temperature scaling
        scorer:  ScorerBase object
                 ScorerBase object of the RelevanceModel
        dataset: RelevanceDataset
                 RelevanceDataset object to be used for training and evaluating temperature scaling
        logger : Logger
                 Logger object to log events
        logs_dir_local: str
                 path to save the TS scores
        temperature: float
                temperature initial value
        file_io:FileIO
                file I/O handler objects for reading and writing data
        zip_output: bool
                boolean value indicates whether the output should be zipped

    Returns
    -------
        `Tuple[np.ndarray, ...]`
        optimizer output containing temperature value learned during temperature scaling
    """
    # get an intermediate model with logits as output --> MUST NOT BE AN ACTIVATION (e.g.
    # SOFTMAX) LAYER
    intermediate_model = get_intermediate_model(model, scorer)

    start = tf.Variable(initial_value=[temperature], shape=(1), dtype=tf.float32, name="temp",
                        trainable=True)

    def nll_loss_fn(temperature_var):
        """Computes Negative Log Likelihood loss by applying temperature scaling.
        Parameters:
            temperature_var: TF.Tensor,  a  parameter of size=(1) to be trained
        Returns:
            NLL loss value (averaged on total examples)
        """
        logits_tensor = tf.constant(logits_numpys, name='logits_tensor', dtype=tf.float32)
        labels_tensor = tf.constant(labels_numpys, name='labels_tensor', dtype=tf.int32)
        logits_w_temp = tf.divide(logits_tensor, temperature_var)

        return tf.reduce_mean(
            tf.nn.sparse_softmax_cross_entropy_with_logits(labels=labels_tensor,
                                                           logits=logits_w_temp))

    @tf.function
    def nll_with_lbfgs():
        """Returns optimizer function. Inspired by
         https://github.com/gpleiss/temperature_scaling/blob/master/temperature_scaling.py#L60"""
        return tfp.optimizer.lbfgs_minimize(make_val_and_grad_fn(nll_loss_fn),
                                            initial_position=start)

    # evaluation on validation set before temperature scaling
    logits_numpys, labels_numpys = get_logits_labels(intermediate_model, dataset.validation)
    original_acc_op, original_nll_loss_op, _ = eval_relevance_model(scorer, logits_numpys,
                                                                    labels_numpys)

    # perform temperature scaling
    results = run_optimizer(nll_with_lbfgs, logger)

    # evaluation on validation set with temperature scaling
    temper = tf.constant(results.position)
    acc_op, _, _ = eval_relevance_model(scorer, logits_numpys, labels_numpys, temperature=temper)

    logger.info("=" * 50)
    logger.info(f'temperature value : {results.position}')
    logger.info('Evaluating on the validation dataset ')
    logger.info(f'original loss: {original_nll_loss_op}, accuracy: {original_acc_op}, \n'
          f'temperature scaling loss: {results.objective_value}, accuracy: {acc_op}\n')
    logger.info("="*50)
    logger.info("Evaluating on the test dataset")

    logits_numpys_test, labels_numpys_test = get_logits_labels(intermediate_model, dataset.test)

    # evaluation on test set before temperature scaling
    acc_test_original, _, original_softmaxes = eval_relevance_model(scorer, logits_numpys_test,
                                                                    labels_numpys_test)

    # evaluation on test set with temperature scaling
    acc_test_temperature_scaling, _, temperature_scaling_softmaxes =\
        eval_relevance_model(scorer, logits_numpys_test, labels_numpys_test, temperature=temper)

    # write the full vector in the csv not ...
    np.set_printoptions(
        formatter={'all': lambda x: str(x.decode('utf-8')) if isinstance(x, bytes) else str(x)},
        linewidth=sys.maxsize,
        threshold=sys.maxsize)
    # avoiding .tolist() that is not memory efficient
    # note: temperature scaling does not change the accuracy as it does not change the maximum. So,
    # the temperature scaling predicted labels must be the same as  original
    # predicted labels: `original_predicted_label`
    data_dic = {'original_scores': [x for x in original_softmaxes.numpy()],
                'temperature_scaling_scores': [x for x in temperature_scaling_softmaxes.numpy()],
                'original_predicted_label': tf.argmax(original_softmaxes, axis=-1).numpy(),
                'true_label': labels_numpys_test,
                }

    logger.info(f'original test accuracy: {acc_test_original}, \ntemperature scaling '
                     f'test accuracy: {acc_test_temperature_scaling} \n')

    # the file is big and can be zipped
    final_dir_path = dict_to_csv(data_dic, logs_dir_local, file_io, zip_output=zip_output)
    logger.info(f"Read and created  {final_dir_path}")
    return results


class TemperatureScalingLayer(tf.keras.layers.Layer):
    """
    Temperature Scaling class , a custom layer. This layer will divide the input tensors by
    temperature value.
    """

    def __init__(self, temperature, name='temperature', **kwargs):
        """
        Constructor method for creating the Temperature Scaling layer

        Parameters
        ----------
            temperature: float
                a float value to be used for initializing the temperature parameter to scale the
                input.
        """
        super(TemperatureScalingLayer, self).__init__(name=name, **kwargs)
        self.temperature = temperature

    def build(self, input_shape):
        """initialize weights for temperature parameter
        Parameters
        ----------
        input_shape: tuple
            a tuple showing layer input shape
        """
        self.temperature_var = self.add_weight("temperature",
                                               shape=[1],
                                               dtype=tf.float32,
                                               initializer=tf.keras.initializers.
                                               Constant(value=self.temperature))

    def get_config(self):
        """
        Returns a Python dict containing the configuration of the layer
        """
        config = super(TemperatureScalingLayer, self).get_config()
        config.update({"temperature": self.temperature})
        return config

    def call(self, input):
        """
        Scales the input and computes input/temperature.
        Parameters
        ---------
        input: `Tensor` object
            Tensors (should be logits) as nominator in scaling operation.

        Returns
        ------
        scaled input w.r.t temperature parameter (input/temperature).
        """
        return tf.divide(input, self.temperature_var)


import os
import numpy as np
import pandas as pd
import random
import tensorflow as tf

from ml4ir.applications.ranking.tests.test_base import RankingTestBase
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.applications.ranking.model.ranking_model import RankingModel
from ml4ir.base.features.feature_config import FeatureConfig


class RankingModelTest(RankingTestBase):
    def run_default_pipeline(self, data_dir: str, data_format: str, feature_config_path: str):
        """Train a model with the default set of args"""
        metrics_keys = ["MRR"]

        # Fix random seed values for repeatability
        tf.keras.backend.clear_session()
        np.random.seed(123)
        tf.random.set_seed(123)
        random.seed(123)

        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=self.args.tfrecord_type,
            feature_config_dict=self.file_io.read_yaml(feature_config_path),
            logger=self.logger,
        )

        relevance_dataset = RelevanceDataset(
            data_dir=data_dir,
            data_format=data_format,
            feature_config=feature_config,
            tfrecord_type=self.args.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns={},
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.file_io,
            logger=self.logger,
        )

        ranking_model: RankingModel = self.get_ranking_model(
            loss_key=self.args.loss_key, feature_config=feature_config, metrics_keys=metrics_keys
        )

        ranking_model.fit(dataset=relevance_dataset, num_epochs=1, models_dir=self.output_dir)

        loss = dict(
            zip(
                ranking_model.model.metrics_names,
                ranking_model.model.evaluate(relevance_dataset.test),
            )
        )["loss"]
        new_MRR = ranking_model.evaluate(
            test_dataset=relevance_dataset.test, logs_dir=self.args.logs_dir,
        )[0]["new_MRR"]

        return loss, new_MRR

    def test_csv_and_tfrecord(self):
        """
        Test model training and evaluate the performance metrics between CSV and TFRecord data
        """

        # Test model training on CSV data
        data_dir = os.path.join(self.root_data_dir, "csv")
        feature_config_path = os.path.join(self.root_data_dir, "configs", self.feature_config_fname)

        csv_loss, csv_mrr = self.run_default_pipeline(
            data_dir=data_dir, data_format="csv", feature_config_path=feature_config_path
        )

        # Check if the loss and accuracy on the test set is the same
        assert np.isclose(csv_loss, 0.56748, rtol=0.01)
        assert np.isclose(csv_mrr, 0.70396, rtol=0.01)

        # Test model training on TFRecord SequenceExample data
        data_dir = os.path.join(self.root_data_dir, "tfrecord")
        feature_config_path = os.path.join(self.root_data_dir, "configs", self.feature_config_fname)

        tfrecord_loss, tfrecord_mrr = self.run_default_pipeline(
            data_dir=data_dir, data_format="tfrecord", feature_config_path=feature_config_path
        )

        # Check if the loss and accuracy on the test set is the same
        assert np.isclose(tfrecord_loss, 0.56748, rtol=0.01)
        assert np.isclose(tfrecord_mrr, 0.70396, rtol=0.01)

        # Compare CSV and TFRecord loss and accuracies
        assert np.isclose(tfrecord_loss, csv_loss, rtol=0.01)
        assert np.isclose(tfrecord_mrr, csv_mrr, rtol=0.01)

    def test_linear_ranking_model_save(self):
        """
        Test the save functionality of LinearRankingModel.
        Specifically, we test to see if the features and coefficients have been saved as CSV file.
        """
        feature_config_path = os.path.join(self.root_data_dir, "configs/linear_model", self.feature_config_fname)
        self.load_model_config(os.path.join(self.root_data_dir, "configs/linear_model", "model_config.yaml"))
        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=self.args.tfrecord_type,
            feature_config_dict=self.file_io.read_yaml(feature_config_path),
            logger=self.logger,
        )

        ranking_model: RankingModel = self.get_ranking_model(
            loss_key=self.args.loss_key,
            feature_config=feature_config,
            metrics_keys=["MRR"]
        )

        # Save the model and check if coefficients file was saved
        ranking_model.save(models_dir=self.args.models_dir)
        assert os.path.exists(os.path.join(self.args.models_dir, "coefficients.csv"))

        # Check coefficients for all features were saved
        coefficients_df = pd.read_csv(
            os.path.join(self.args.models_dir, "coefficients.csv"))
        train_features = set(feature_config.get_train_features("node_name"))

        # Adding +1 to account for bias term
        assert len(train_features) + 1 == coefficients_df.shape[0]
        for train_feature in train_features:
            assert train_feature in coefficients_df.feature.values




import os
import numpy as np
from tensorflow.keras import models as kmodels
import tensorflow as tf

from ml4ir.applications.ranking.tests.test_base import RankingTestBase
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.config.keys import DataFormatKey
from ml4ir.applications.ranking.model.ranking_model import RankingModel
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import ServingSignatureKey


class RankingModelTest(RankingTestBase):
    def test_model_serving(self):
        """
        Train a simple model and test serving flow by loading the SavedModel
        """

        # Test model training on TFRecord SequenceExample data
        data_dir = os.path.join(self.root_data_dir, "tfrecord")
        self.feature_config_fname = "feature_config_integration_test.yaml"
        feature_config: FeatureConfig = self.get_feature_config()

        metrics_keys = ["categorical_accuracy"]

        def get_dataset(parse_tfrecord):
            return RelevanceDataset(
                data_dir=data_dir,
                data_format=DataFormatKey.TFRECORD,
                feature_config=feature_config,
                tfrecord_type=self.args.tfrecord_type,
                max_sequence_size=self.args.max_sequence_size,
                batch_size=self.args.batch_size,
                preprocessing_keys_to_fns={},
                train_pcent_split=self.args.train_pcent_split,
                val_pcent_split=self.args.val_pcent_split,
                test_pcent_split=self.args.test_pcent_split,
                use_part_files=self.args.use_part_files,
                parse_tfrecord=parse_tfrecord,
                file_io=self.file_io,
                logger=self.logger,
            )

        # Get raw TFRecord dataset
        raw_dataset = get_dataset(parse_tfrecord=False)

        # Parse the raw TFRecord dataset
        parsed_dataset = get_dataset(parse_tfrecord=True)

        model: RankingModel = self.get_ranking_model(
            loss_key=self.args.loss_key, feature_config=feature_config, metrics_keys=metrics_keys
        )

        model.fit(dataset=parsed_dataset, num_epochs=1, models_dir=self.output_dir)

        model.save(
            models_dir=self.args.models_dir,
            preprocessing_keys_to_fns={},
            postprocessing_fn=None,
            required_fields_only=not self.args.use_all_fields_at_inference,
            pad_sequence=self.args.pad_sequence_at_inference,
        )

        # Load SavedModel and get the right serving signature
        default_model = kmodels.load_model(
            os.path.join(self.output_dir, "final", "default"), compile=False
        )
        assert ServingSignatureKey.DEFAULT in default_model.signatures
        default_signature = default_model.signatures[ServingSignatureKey.DEFAULT]

        tfrecord_model = kmodels.load_model(
            os.path.join(self.output_dir, "final", "tfrecord"), compile=False
        )
        assert ServingSignatureKey.TFRECORD in tfrecord_model.signatures
        tfrecord_signature = tfrecord_model.signatures[ServingSignatureKey.TFRECORD]

        # Fetch a single batch for testing
        sequence_example_protos = next(iter(raw_dataset.test))
        parsed_sequence_examples = next(iter(parsed_dataset.test))[0]
        parsed_dataset_batch = parsed_dataset.test.take(1)

        # Use the loaded serving signatures for inference
        model_predictions = model.predict(parsed_dataset_batch)[self.args.output_name].values
        default_signature_predictions = default_signature(**parsed_sequence_examples)[
            self.args.output_name
        ]

        # Since we do not pad dummy records in tfrecord serving signature,
        # we can only predict on a single record at a time
        tfrecord_signature_predictions = [
            tfrecord_signature(protos=tf.gather(sequence_example_protos, [i]))[
                self.args.output_name
            ]
            for i in range(self.args.batch_size)
        ]

        def _flatten_records(x):
            """Collapse first two dimensions of a tensor -> [batch_size, max_num_records]"""
            return tf.reshape(x, tf.concat([[-1], tf.shape(x)[2:]], axis=0))

        def _filter_records(x, mask):
            """
            Filter records that were padded in each query
            Input shape: [batch_size, num_features]
            Output shape: [batch_size, num_features]
            """
            return tf.squeeze(tf.gather_nd(x, tf.where(tf.not_equal(mask, 0))))

        # Get mask for padded values
        mask = _flatten_records(parsed_sequence_examples["mask"])

        # Flatten scores to each record and filter out scores from padded records
        default_signature_predictions = _filter_records(
            _flatten_records(default_signature_predictions), mask
        )
        tfrecord_signature_predictions = tf.squeeze(
            tf.concat(tfrecord_signature_predictions, axis=1)
        )

        # Compare the scores from the different versions of the model
        assert np.isclose(model_predictions, default_signature_predictions, rtol=0.01,).all()

        assert np.isclose(model_predictions, tfrecord_signature_predictions, rtol=0.01,).all()

        assert np.isclose(
            default_signature_predictions, tfrecord_signature_predictions, rtol=0.01,
        ).all()

    def get_feature_config(self):
        feature_config_path = os.path.join(
            self.root_data_dir, "configs", self.feature_config_fname
        )

        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=self.args.tfrecord_type,
            feature_config_dict=self.file_io.read_yaml(feature_config_path),
            logger=self.logger,
        )

        return feature_config

    def get_tfrecord_signature(self, feature_config: FeatureConfig):
        metrics_keys = ["categorical_accuracy"]
        model: RankingModel = self.get_ranking_model(
            loss_key=self.args.loss_key, feature_config=feature_config, metrics_keys=metrics_keys
        )
        model.save(
            models_dir=self.args.models_dir,
            preprocessing_keys_to_fns={},
            postprocessing_fn=None,
            required_fields_only=not self.args.use_all_fields_at_inference,
            pad_sequence=self.args.pad_sequence_at_inference,
        )

        # Load SavedModel and get the right serving signature
        tfrecord_model = kmodels.load_model(
            os.path.join(self.output_dir, "final", "tfrecord"), compile=False
        )
        assert ServingSignatureKey.TFRECORD in tfrecord_model.signatures

        return tfrecord_model.signatures[ServingSignatureKey.TFRECORD]

    def test_serving_n_records(self):
        """Test serving signature with different number of records"""
        feature_config: FeatureConfig = self.get_feature_config()
        tfrecord_signature = self.get_tfrecord_signature(feature_config)

        for num_records in range(1, 120):
            proto = tf.constant(
                [feature_config.create_dummy_protobuf(num_records=num_records).SerializeToString()]
            )
            try:
                tfrecord_signature(protos=proto)
            except Exception:
                assert False

    def test_serving_required_fields_only(self):
        """Test serving signature with protos with only required fields"""
        feature_config: FeatureConfig = self.get_feature_config()
        tfrecord_signature = self.get_tfrecord_signature(feature_config)

        proto = tf.constant(
            [feature_config.create_dummy_protobuf(required_only=True).SerializeToString()]
        )

        try:
            tfrecord_signature(protos=proto)
        except Exception:
            assert False



import ast
import argparse
import pandas as pd
import numpy as np

max_f_id = 0
def process_line(line, keep_additional_info, query_id_name, relevance_name):
    """Takes an input line in ranklib format and returns a row in ml4ir format.

        Parameters
        ----------
        line : str
            a line from ranklib format data
        keep_additional_info : bool
            Option to keep additional info (All info after the "#") True to keep, False to ignore.
        query_id_name : str
            The name of the query id column.
        relevance_name : str
            The name of the relevance column (the target label).


        Returns
        -------
        dictionary <column:value>
            Keys are the column names values are the parsed values.
    """

    if keep_additional_info:
        feature_values = line.replace('#', '').replace(' = ', ':').strip().split()
    else:
        feature_values = line.split('#')[0].strip().split()
    feature_values[0] = relevance_name + ':' + feature_values[0]
    r = {}
    for fv in feature_values:
        feat = fv.split(':')[0].strip()
        if feat == query_id_name:
            val = fv.split(':')[1].strip()
            r[feat] = 'Q'+str(val)
        elif feat == relevance_name:
            val = fv.split(':')[1].strip()
            r[feat] = float(val)
        else:
            try:
                val = float(fv.split(':')[1].strip())
            except:
                val = fv.split(':')[1].strip()
            r['f_' + feat] = val
            global max_f_id
            if int(feat) > max_f_id:
                max_f_id = int(feat)
    return r


def convert(input_file, keep_additional_info, gl_2_clicks,
            non_zero_features_only, query_id_name, relevance_name,
            add_dummy_rank_column = True):
    """Convert the input file with the specified parameters into ml4ir format. returns a dataframe

            Parameters
            ----------
            input_file : str
                ranklib input file path
            keep_additional_info : bool
                Option to keep additional info (All info after the "#") True to keep, False to ignore.
            gl_2_clicks : int
                Convert graded relevance to clicks (only max relevant document is considered clicked) 1 to convert
            query_id_name : str
                The name of the query id column.
            relevance_name : str
                The name of the relevance column (the target label).
            add_dummy_rank_column : bool
                ml4ir expects pre-rankings. This would add a dummy pre-rankings.


            Returns
            -------
            Dataframe
                converted ml4ir dataframe
        """

    f = open(input_file, 'r')
    rows = []
    for line in f:
        rows.append(process_line(line, keep_additional_info, query_id_name, relevance_name))
    f.close()
    if non_zero_features_only:
        columns = [query_id_name, relevance_name] + ['f_' + str(i) for i in range(max_f_id)]
        df = pd.DataFrame(rows, columns=columns)
        df.replace(np.nan, 0, inplace=True)
    else:
        df = pd.DataFrame(rows)
    if int(gl_2_clicks) == 1:
        groups = df.groupby(query_id_name)
        for gname, group in groups:
            df.loc[df[query_id_name] == gname, relevance_name] = (
                    df.loc[df[query_id_name] == gname].relevance == max(group.relevance)).astype(int)

    # NOTE: ml4ir expects a pre-ranking. Adding a dummy pre-ranking to match format.
    if add_dummy_rank_column:
        df['rank'] = 1
    return df

def ranklib_to_csv(input_file, output_file, keep_additional_info, gl_2_clicks, non_zero_features_only, query_id_name, relevance_name, add_dummy_rank_column = False):
    """Convert the input file with the specified parameters into ml4ir format writes the converted file to a csv

            Parameters
            ----------
            input_file : str
                ranklib input file path
            output_file : str
                output converted file path
            keep_additional_info : bool
                Option to keep additional info (All info after `#`) True to keep, False to ignore.
            gl_2_clicks : int
                Convert graded relevance to clicks (only max relevant document is considered clicked) 1 to convert
            query_id_name : str
                The name of the query id column.
            relevance_name : str
                The name of the relevance column (the target label).
            add_dummy_rank_column : bool
                ml4ir expects pre-rankings. This would add a dummy pre-rankings.
    """

    df = convert(input_file, keep_additional_info, gl_2_clicks, non_zero_features_only, query_id_name, relevance_name, add_dummy_rank_column)
    df.to_csv(output_file)

def ranklib_directory_to_csvs(input_dir, keep_additional_info, gl_2_clicks, non_zero_features_only, query_id_name, relevance_name, add_dummy_rank_column = False):
    """Convert all files in the given directory with the specified parameters into ml4ir format writes the converted file to a csv
            Parameters
            ----------
            input_dir : str
                ranklib input directory path. All files within the directory will be converted.
            keep_additional_info : bool
                Option to keep additional info (All info after `#`) True to keep, False to ignore.
            gl_2_clicks : int
                Convert graded relevance to clicks (only max relevant document is considered clicked) 1 to convert
            query_id_name : str
                The name of the query id column.
            relevance_name : str
                The name of the relevance column (the target label).
            add_dummy_rank_column : bool
                ml4ir expects pre-rankings. This would add a dummy pre-rankings.
    """
    from os import listdir
    from os.path import isfile, join
    onlyfiles = [f for f in listdir(input_dir) if isfile(join(input_dir, f))]
    for f in onlyfiles[1:]:
        ranklib_to_csv(join(input_dir, f), join(input_dir, f)+'_ml4ir.csv', keep_additional_info, gl_2_clicks, non_zero_features_only, query_id_name, relevance_name, add_dummy_rank_column)



if __name__ == "__main__":
    # parsing arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--input_file', type=str, default='ml4ir/applications/ranking/tests/data/train/sample.txt', help='ranklib input file path')
    parser.add_argument('--input_dir', type=str, default='ml4ir/applications/ranking/tests/data/test',
                        help='ranklib input directory path. All files within the directory will be converted.')
    parser.add_argument('--output_file', type=str, default='ml4ir/applications/ranking/tests/data/train/sample_ml4ir.csv', help='output converted file path')
    parser.add_argument('--keep_additional_info', type=ast.literal_eval, default=True,
                        help='Option to keep additional info (All info after the "#") True to keep, False to ignore')
    parser.add_argument('--gl_2_clicks', type=int, default=1,
                        help='Convert graded relevance to clicks (only max relevant document is considered clicked) 1 to convert')
    parser.add_argument('--non_zero_features_only', type=int, default=True,
                        help='Only non zero features are stored. True for yes, False otherwise')
    parser.add_argument('--query_id_name', type=str, default='qid',
                        help='The name of the query id column.')
    parser.add_argument('--relevance_name', type=str, default='relevance',
                        help='The name of the relevance column.')

    args = parser.parse_args()
    print("Converting file...")
    ranklib_to_csv(args.input_file, args.output_file, args.keep_additional_info, args.gl_2_clicks,
                   args.non_zero_features_only, args.query_id_name, args.relevance_name)
    print('Conversion is completed')



from tensorflow.keras import losses
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.losses import Reduction

from ml4ir.applications.ranking.model.losses.loss_base import PointwiseLossBase


class SigmoidCrossEntropy(PointwiseLossBase):
    def get_loss_fn(self, **kwargs):
        """
        Define a sigmoid cross entropy loss
        Additionally can pass in record positions to handle positional bias

        Returns
        -------
        function
            Function to compute sigmoid cross entropy loss

        Notes
        -----
            Uses `mask` field to exclude padded records from contributing
            to the loss
        """
        bce = losses.BinaryCrossentropy(reduction=Reduction.SUM_OVER_BATCH_SIZE)
        mask = kwargs.get("mask")

        def _loss_fn(y_true, y_pred):
            """
            Shapes
            ------
            y_true : [batch_size, num_classes]
            y_pred : [batch_size, num_classes]
            mask : [batch_size, num_classes]
            """
            # Mask the padded records
            y_true = tf.gather_nd(y_true, tf.where(tf.equal(mask, tf.constant(1.0))))
            y_pred = tf.gather_nd(y_pred, tf.where(tf.equal(mask, tf.constant(1.0))))

            return bce(y_true, y_pred)

        return _loss_fn

    def get_final_activation_op(self, output_name):
        """
        Define a sigmoid activation function

        Parameters
        ----------
        output_name : str
            Name of the output score node in the tensorflow model

        Returns
        -------
        function
            Function to apply sigmoid activation to the output score
        """
        return lambda logits, mask: layers.Activation("sigmoid", name=output_name)(logits)



"""Cyclical Learning Rate Schedule policies for TensorFlow.
Adopted from:
https://www.tensorflow.org/addons/api_docs/python/tfa/optimizers/CyclicalLearningRate
"""

import tensorflow as tf
from typing import Union, Callable


class CyclicalLearningRate(tf.keras.optimizers.schedules.LearningRateSchedule):
    """A LearningRateSchedule that uses cyclical schedule."""

    def __init__(
        self,
        initial_learning_rate,
        maximal_learning_rate,
        step_size,
        scale_fn: Callable,
        scale_mode: str = "cycle",
        name: str = "CyclicalLearningRate",
    ):
        """Applies cyclical schedule to the learning rate.
        See Cyclical Learning Rates for Training Neural Networks. https://arxiv.org/abs/1506.01186
        ```python
        lr_schedule = tf.keras.optimizers.schedules.CyclicalLearningRate(
            initial_learning_rate=1e-4,
            maximal_learning_rate=1e-2,
            step_size=2000,
            scale_fn=lambda x: 1.,
            scale_mode="cycle",
            name="MyCyclicScheduler")
        model.compile(optimizer=tf.keras.optimizers.SGD(
                                                    learning_rate=lr_schedule),
                      loss='sparse_categorical_crossentropy',
                      metrics=['accuracy'])
        model.fit(data, labels, epochs=5)
        ```
        You can pass this schedule directly into a
        `tf.keras.optimizers.Optimizer` as the learning rate.

        Parameters
        ----------
            initial_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The initial learning rate.
            maximal_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The maximum learning rate.
            step_size: A scalar `float32` or `float64` `Tensor` or a
                Python number. Step size.
            scale_fn: A function. Scheduling function applied in cycle
            scale_mode: ['cycle', 'iterations']. Mode to apply during cyclic
                schedule
            name: (Optional) Name for the operation.


        Returns
        -------
            Updated learning rate value.
        """
        super().__init__()
        self.initial_learning_rate = initial_learning_rate
        self.maximal_learning_rate = maximal_learning_rate
        self.step_size = step_size
        self.scale_fn = scale_fn
        self.scale_mode = scale_mode
        self.name = name

    def __call__(self, step):
        with tf.name_scope(self.name or "CyclicalLearningRate"):
            initial_learning_rate = tf.convert_to_tensor(
                self.initial_learning_rate, name="initial_learning_rate"
            )
            dtype = initial_learning_rate.dtype
            maximal_learning_rate = tf.cast(self.maximal_learning_rate, dtype)
            step_size = tf.cast(self.step_size, dtype)
            cycle = tf.floor(1 + step / (2 * step_size))
            x = tf.abs(step / step_size - 2 * cycle + 1)

            mode_step = cycle if self.scale_mode == "cycle" else step

            lr = initial_learning_rate + (
                maximal_learning_rate - initial_learning_rate
            ) * tf.maximum(tf.cast(0, dtype), (1 - x)) * self.scale_fn(mode_step)

            return lr

    def get_config(self):
        return {
            "initial_learning_rate": self.initial_learning_rate,
            "maximal_learning_rate": self.maximal_learning_rate,
            "scale_fn": self.scale_fn,
            "step_size": self.step_size,
            "scale_mode": self.scale_mode,
        }


class TriangularCyclicalLearningRate(CyclicalLearningRate):
    def __init__(
        self,
        initial_learning_rate,
        maximal_learning_rate,
        step_size,
        scale_mode: str = "cycle",
        name: str = "TriangularCyclicalLearningRate",
    ):
        """Applies triangular cyclical schedule to the learning rate.
        See Cyclical Learning Rates for Training Neural Networks. https://arxiv.org/abs/1506.01186
        ```python
        from tf.keras.optimizers import schedules
        lr_schedule = schedules.TriangularCyclicalLearningRate(
            initial_learning_rate=1e-4,
            maximal_learning_rate=1e-2,
            step_size=2000,
            scale_mode="cycle",
            name="MyCyclicScheduler")
        model.compile(optimizer=tf.keras.optimizers.SGD(
                                                    learning_rate=lr_schedule),
                      loss='sparse_categorical_crossentropy',
                      metrics=['accuracy'])
        model.fit(data, labels, epochs=5)
        ```
        You can pass this schedule directly into a
        `tf.keras.optimizers.Optimizer` as the learning rate.
        Args:
            initial_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The initial learning rate.
            maximal_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The maximum learning rate.
            step_size: A scalar `float32` or `float64` `Tensor` or a
                Python number. Step size.
            scale_mode: ['cycle', 'iterations']. Mode to apply during cyclic
                schedule
            name: (Optional) Name for the operation.
        Returns:
            Updated learning rate value.
        """
        super().__init__(
            initial_learning_rate=initial_learning_rate,
            maximal_learning_rate=maximal_learning_rate,
            step_size=step_size,
            scale_fn=lambda x: 1.0,
            scale_mode=scale_mode,
            name=name,
        )

    def get_config(self):
        return {
            "initial_learning_rate": self.initial_learning_rate,
            "maximal_learning_rate": self.maximal_learning_rate,
            "step_size": self.step_size,
            "scale_mode": self.scale_mode,
        }


class Triangular2CyclicalLearningRate(CyclicalLearningRate):
    #@typechecked
    def __init__(
        self,
        initial_learning_rate,
        maximal_learning_rate,
        step_size,
        scale_mode: str = "cycle",
        name: str = "Triangular2CyclicalLearningRate",
    ):
        """Applies triangular2 cyclical schedule to the learning rate.
        See Cyclical Learning Rates for Training Neural Networks. https://arxiv.org/abs/1506.01186
        ```python
        from tf.keras.optimizers import schedules
        lr_schedule = schedules.Triangular2CyclicalLearningRate(
            initial_learning_rate=1e-4,
            maximal_learning_rate=1e-2,
            step_size=2000,
            scale_mode="cycle",
            name="MyCyclicScheduler")
        model.compile(optimizer=tf.keras.optimizers.SGD(
                                                    learning_rate=lr_schedule),
                      loss='sparse_categorical_crossentropy',
                      metrics=['accuracy'])
        model.fit(data, labels, epochs=5)
        ```
        You can pass this schedule directly into a
        `tf.keras.optimizers.Optimizer` as the learning rate.
        Args:
            initial_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The initial learning rate.
            maximal_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The maximum learning rate.
            step_size: A scalar `float32` or `float64` `Tensor` or a
                Python number. Step size.
            scale_mode: ['cycle', 'iterations']. Mode to apply during cyclic
                schedule
            name: (Optional) Name for the operation.
        Returns:
            Updated learning rate value.
        """
        super().__init__(
            initial_learning_rate=initial_learning_rate,
            maximal_learning_rate=maximal_learning_rate,
            step_size=step_size,
            scale_fn=lambda x: 1 / (2.0 ** (x - 1)),
            scale_mode=scale_mode,
            name=name,
        )

    def get_config(self):
        return {
            "initial_learning_rate": self.initial_learning_rate,
            "maximal_learning_rate": self.maximal_learning_rate,
            "step_size": self.step_size,
            "scale_mode": self.scale_mode,
        }


class ExponentialCyclicalLearningRate(CyclicalLearningRate):
    def __init__(
        self,
        initial_learning_rate,
        maximal_learning_rate,
        step_size,
        gamma=1.0,
        scale_mode: str = "iterations",
        name: str = "ExponentialCyclicalLearningRate",
    ):
        """Applies exponential cyclical schedule to the learning rate.
        See Cyclical Learning Rates for Training Neural Networks. https://arxiv.org/abs/1506.01186
        ```python
        from tf.keras.optimizers import schedules
        lr_schedule = ExponentialCyclicalLearningRate(
            initial_learning_rate=1e-4,
            maximal_learning_rate=1e-2,
            step_size=2000,
            scale_mode="cycle",
            gamma=0.96,
            name="MyCyclicScheduler")
        model.compile(optimizer=tf.keras.optimizers.SGD(
                                                    learning_rate=lr_schedule),
                      loss='sparse_categorical_crossentropy',
                      metrics=['accuracy'])
        model.fit(data, labels, epochs=5)
        ```
        You can pass this schedule directly into a
        `tf.keras.optimizers.Optimizer` as the learning rate.
        Args:
            initial_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The initial learning rate.
            maximal_learning_rate: A scalar `float32` or `float64` `Tensor` or
                a Python number.  The maximum learning rate.
            step_size: A scalar `float32` or `float64` `Tensor` or a
                Python number. Step size.
            scale_mode: ['cycle', 'iterations']. Mode to apply during cyclic
                schedule
            gamma: A scalar `float32` or `float64` `Tensor` or a
                Python number.  Gamma value.
            name: (Optional) Name for the operation.
        Returns:
            Updated learning rate value.
        """
        self.gamma = gamma
        super().__init__(
            initial_learning_rate=initial_learning_rate,
            maximal_learning_rate=maximal_learning_rate,
            step_size=step_size,
            scale_fn=lambda x: gamma ** x,
            scale_mode=scale_mode,
            name=name,
        )

    def get_config(self):
        return {
            "initial_learning_rate": self.initial_learning_rate,
            "maximal_learning_rate": self.maximal_learning_rate,
            "step_size": self.step_size,
            "scale_mode": self.scale_mode,
            "gamma": self.gamma,
        }


import unittest
import warnings
import pandas as pd
import numpy as np
import pathlib
import tensorflow as tf
from testfixtures import TempDirectory
from ml4ir.applications.ranking.pipeline import RankingPipeline
from ml4ir.applications.ranking.config.parse_args import get_args

warnings.filterwarnings("ignore")


def create_parse_args(data_dir, feature_config, model_config, logs_dir, num_folds, use_testset_in_folds):
    argv = ["--data_dir", data_dir,
            "--feature_config", feature_config,
            "--kfold", str(num_folds),
            "--include_testset_in_kfold", str(use_testset_in_folds),
            "--run_id", "testing_kfold_cs",
            "--data_format", "csv",
            "--execution_mode", "train_inference_evaluate",
            "--num_epochs", "1",
            "--model_config", model_config,
            "--batch_size", "4",
            "--logs_dir", logs_dir,
            "--max_sequence_size", "25",
            "--train_pcent_split", "1.0",
            "--val_pcent_split", "-1",
            "--test_pcent_split", "-1"]

    args = get_args(argv)
    return args


class TestML4IRKfoldCV(unittest.TestCase):
    """
    Test kfold cross validation
    """

    def setUp(self):
        self.dir = pathlib.Path(__file__).parent
        self.working_dir = TempDirectory()
        self.log_dir = self.working_dir.makedir('logs')
        self.working_dir.makedir('train')
        self.working_dir.makedir('test')
        self.working_dir.makedir('validation')
        tf.random.set_seed(123)
        np.random.seed(123)

    def tearDown(self):
        TempDirectory.cleanup_all()

    def setup_data(self, dataset_name, num_features, num_folds, use_testset_in_folds):
        """
        reading dataset and splitting it into 3 groups: train, validation and test.
        Parameters
        ----------
        dataset_name: str
            dataset to load (from the sanity tests datasets)
        num_features: int
            number of features in the dataset
        num_folds: int
            number of folds
        use_testset_in_folds: bool
            whether to include the testset in the merge
        """
        working_dir = pathlib.Path(self.working_dir.path)
        log_dir = pathlib.Path(self.log_dir)

        df = pd.read_csv(pathlib.Path(__file__).parent / "data" / "linear_sanity_tests" / dataset_name)
        fconfig_name = "feature_config_sanity_tests_" + str(num_features) + "_features.yaml"
        feature_config_file = pathlib.Path(__file__).parent / "data" / "configs" / fconfig_name
        model_config_file = pathlib.Path(__file__).parent / "data" / "configs" / "model_config_sanity_tests.yaml"

        queries = df.groupby('query_id')
        dfs = np.array_split(queries, 3)
        pd.concat([dfs[0][i][-1] for i in range(len(dfs[0]))]).to_csv(working_dir / 'train' / 'data.csv')
        pd.concat([dfs[1][i][-1] for i in range(len(dfs[1]))]).to_csv(working_dir / 'validation' / 'data.csv')
        pd.concat([dfs[2][i][-1] for i in range(len(dfs[2]))]).to_csv(working_dir / 'test' / 'data.csv')
        return create_parse_args(data_dir=working_dir.as_posix(),
                                 feature_config=feature_config_file.as_posix(),
                                 model_config=model_config_file.as_posix(),
                                 logs_dir=log_dir.as_posix(),
                                 num_folds=num_folds,
                                 use_testset_in_folds=use_testset_in_folds)

    def run_merge_datasets_test(self, dataset_name, num_features, num_folds, use_testset_in_folds, expected_num_queries):
        """
        With the specified input parameters, merge the datasets together and compare against the expected number of
        queries.

        Parameters
        ----------
        dataset_name: str
            dataset to load (from the sanity tests datasets)
        num_features: int
            number of features in the dataset
        num_folds: int
            number of folds
        use_testset_in_folds: bool
            whether to include the testset in the merge
        expected_num_queries: int
            expected number of queries ion the final merge.
        """
        args = self.setup_data(dataset_name, num_features, num_folds, use_testset_in_folds)

        rp = RankingPipeline(args=args)
        relevance_dataset = rp.get_kfold_relevance_dataset(args.kfold, args.include_testset_in_kfold,
                                                           read_data_sets=True)

        all_data = relevance_dataset.merge_datasets()
        query_ids = set([q[0]['query_id'].numpy()[0] for q in all_data])
        assert len(query_ids) == expected_num_queries

    def run_folds_creation_test(self, dataset_name, num_features, num_folds, use_testset_in_folds):
        """
        Read and merge datasets, then create folds. A successful fold creation should have different, non
        overlapping train, validation and test sets.

        Parameters
        ----------
        dataset_name: str
            dataset to load (from the sanity tests datasets)
        num_features: int
            number of features in the dataset
        num_folds: int
            number of folds
        use_testset_in_folds: bool
            whether to include the testset in the merge
        """
        args = self.setup_data(dataset_name, num_features, num_folds, use_testset_in_folds)

        rp = RankingPipeline(args=args)
        relevance_dataset = rp.get_kfold_relevance_dataset(args.kfold, args.include_testset_in_kfold,
                                                           read_data_sets=True)

        all_data = relevance_dataset.merge_datasets()
        for i in range(num_folds):
            fold_relevance_dataset = rp.get_kfold_relevance_dataset(args.kfold, args.include_testset_in_kfold,
                                                               read_data_sets=False)
            fold_relevance_dataset.create_folds(i, all_data, relevance_dataset)
            train = fold_relevance_dataset.train.unbatch()
            validation = fold_relevance_dataset.validation.unbatch()
            if use_testset_in_folds:
                test = fold_relevance_dataset.test.unbatch()
            train_qids = set([q[0]['query_id'].numpy()[0] for q in train])
            validation_qids = set([q[0]['query_id'].numpy()[0] for q in validation])
            if use_testset_in_folds:
                test_qids = set([q[0]['query_id'].numpy()[0] for q in test])
            if use_testset_in_folds:
                assert len(set.intersection(train_qids, test_qids)) == 0
                assert len(set.intersection(validation_qids, test_qids)) == 0
            assert len(set.intersection(train_qids, validation_qids)) == 0

    def test_merge_datasets_1(self):
        """
        Testing the dataset merge function with testset included in the merge.
        """
        dataset_name = "dataset1.csv"
        num_features = 2
        use_testset_in_folds = True
        expected_num_queries = 41
        num_folds = 5
        self.run_merge_datasets_test(dataset_name, num_features, num_folds, use_testset_in_folds, expected_num_queries)

    def test_merge_datasets_2(self):
        """
        Testing the dataset merge function without including testset in the merge
        """
        dataset_name = "dataset1.csv"
        num_features = 2
        use_testset_in_folds = False
        expected_num_queries = 28
        num_folds = 5
        self.run_merge_datasets_test(dataset_name, num_features, num_folds, use_testset_in_folds, expected_num_queries)

    def test_folds_creation_1(self):
        """
        Testing creating folds. Train, validation and test sets should be different with no over lap
        """
        dataset_name = "dataset1.csv"
        num_features = 2
        use_testset_in_folds = True
        num_folds = 5
        self.run_folds_creation_test(dataset_name, num_features, num_folds, use_testset_in_folds)

    def test_folds_creation_2(self):
        """
        Testing creating folds. Train, validation and test sets should be different with no over lap
        """
        dataset_name = "dataset1.csv"
        num_features = 2
        use_testset_in_folds = False
        num_folds = 5
        self.run_folds_creation_test(dataset_name, num_features, num_folds, use_testset_in_folds)

    def test_folds_creation_3(self):
        """
        Testing creating folds. Train, validation and test sets should be different with no over lap
        """
        dataset_name = "dataset1.csv"
        num_features = 2
        use_testset_in_folds = False
        num_folds = 10
        self.run_folds_creation_test(dataset_name, num_features, num_folds, use_testset_in_folds)

    def test_folds_creation_4(self):
        """
        Testing creating folds. Train, validation and test sets should be different with no over lap
        """
        dataset_name = "dataset1.csv"
        num_features = 2
        use_testset_in_folds = False
        num_folds = 10
        self.run_folds_creation_test(dataset_name, num_features, num_folds, use_testset_in_folds)

    def test_folds_creation_5(self):
        """
        Testing creating folds. Train, validation and test sets should be different with no over lap
        """
        dataset_name = "dataset2.csv"
        num_features = 2
        use_testset_in_folds = True
        num_folds = 10
        self.run_folds_creation_test(dataset_name, num_features, num_folds, use_testset_in_folds)

    def test_folds_creation_6(self):
        """
        Testing creating folds. Train, validation and test sets should be different with no over lap
        """
        dataset_name = "dataset2.csv"
        num_features = 2
        use_testset_in_folds = False
        num_folds = 10
        self.run_folds_creation_test(dataset_name, num_features, num_folds, use_testset_in_folds)


if __name__ == "__main__":
    unittest.main()



import tensorflow as tf
from tensorflow.keras import layers
from typing import List

from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.features.feature_fns.categorical import get_vocabulary_info
from ml4ir.applications.ranking.config.keys import PositionalBiasHandler
from ml4ir.base.io.file_io import FileIO
from ml4ir.base.model.architectures.fixed_additive_positional_bias import FixedAdditivePositionalBias


OOV = 1



class DNNLayer:
    DENSE = "dense"
    BATCH_NORMALIZATION = "batch_norm"
    DROPOUT = "dropout"
    ACTIVATION = "activation"


class DNN:
    def __init__(self, model_config: dict, feature_config: FeatureConfig, file_io):
        self.file_io: FileIO = file_io
        self.model_config = model_config
        self.feature_config = feature_config
        self.layer_ops: List = self.define_architecture(model_config, feature_config)
        if 'positional_bias_handler' in self.model_config and self.model_config['positional_bias_handler'][
            'key'] == PositionalBiasHandler.FIXED_ADDITIVE_POSITIONAL_BIAS:
            self.positional_bias_layer = FixedAdditivePositionalBias(max_ranks=self.model_config['positional_bias_handler']['max_ranks'],
                                                                     kernel_initializer=self.model_config['positional_bias_handler'].get('kernel_initializer', 'Zeros'),
                                                                     l1_coeff=self.model_config['positional_bias_handler'].get('l1_coeff', 0),
                                                                     l2_coeff=self.model_config['positional_bias_handler'].get('l2_coeff', 0))

    def define_architecture(self, model_config: dict, feature_config: FeatureConfig):
        """
        Convert the model from model_config to a List of tensorflow.keras.layer

        :param model_config: dict corresponding to the model config
        :param feature_config: dict corresponding to the feature config, only used in case of classification if the last
            layer of the model_config doesn't have a units number defined (or set to -1). In which case we retrieve the
            label vocabulary defined in the feature_config to deduce the number of units.
        :return: List[layers]: list of keras layer corresponding to each of the layers defined in the model_config.
        """
        def get_op(layer_type, layer_args):
            if layer_type == DNNLayer.DENSE:
                if not "units" in layer_args or layer_args["units"] == -1:
                    try:
                        label_feature_info = feature_config.get_label()
                        vocabulary_keys, vocabulary_ids = get_vocabulary_info(label_feature_info, self.file_io)
                        layer_args["units"] = len(vocabulary_keys) + OOV
                    except:
                        raise KeyError("We were not able to find information for the output layer of your DNN. "
                                       "Try specifying the number of output units either by passing \"units\" in the "
                                       "model configuration yaml file or units in the feature configuration file.")
                return layers.Dense(**layer_args)
            elif layer_type == DNNLayer.BATCH_NORMALIZATION:
                return layers.BatchNormalization(**layer_args)
            elif layer_type == DNNLayer.DROPOUT:
                return layers.Dropout(**layer_args)
            elif layer_type == DNNLayer.ACTIVATION:
                return layers.Activation(**layer_args)
            else:
                raise KeyError("Layer type is not supported : {}".format(layer_type))

        return [
            get_op(layer_args["type"], {k: v for k, v in layer_args.items() if k not in "type"})
            for layer_args in model_config["layers"]
        ]

    def get_architecture_op(self):
        def _architecture_op(ranking_features, metadata_features):
            layer_input = ranking_features

            # Pass ranking features through all the layers of the DNN
            for layer_op in self.layer_ops:
                layer_input = layer_op(layer_input)

            if 'positional_bias_handler' in self.model_config and \
                    self.model_config['positional_bias_handler']['key'] == \
                    PositionalBiasHandler.FIXED_ADDITIVE_POSITIONAL_BIAS:

                positional_bias = self.positional_bias_layer(metadata_features[self.feature_config.get_rank()['node_name']])
                layer_input = layers.Add()([positional_bias, layer_input])

            # Collapse extra dimensions
            if isinstance(self.layer_ops[-1], layers.Dense) and (self.layer_ops[-1].units == 1):
                scores = tf.squeeze(layer_input, axis=-1)
            else:
                scores = layer_input

            return scores

        return _architecture_op
